"""Unit tests for CurveCryptoPool"""
import boa
from hypothesis import HealthCheck, assume, given, settings
from hypothesis import strategies as st

from curvesim.pool import CurveCryptoPool
from curvesim.pool.cryptoswap.calcs import get_y, halfpow, newton_D
from curvesim.pool.cryptoswap.calcs.factory_2_coin import (
    MAX_A,
    MAX_GAMMA,
    MIN_A,
    MIN_GAMMA,
    PRECISION,
    _sqrt_int,
    geometric_mean,
)


def initialize_pool(vyper_cryptopool):
    """
    Initialize python-based pool from the state variables of the
    vyper-based implementation.
    """
    A = vyper_cryptopool.A()
    gamma = vyper_cryptopool.gamma()
    n_coins = vyper_cryptopool.N_COINS()
    precisions = vyper_cryptopool.eval("self._get_precisions()")
    mid_fee = vyper_cryptopool.mid_fee()
    out_fee = vyper_cryptopool.out_fee()
    allowed_extra_profit = vyper_cryptopool.allowed_extra_profit()
    fee_gamma = vyper_cryptopool.fee_gamma()
    adjustment_step = vyper_cryptopool.adjustment_step()
    admin_fee = vyper_cryptopool.admin_fee()
    ma_half_time = vyper_cryptopool.ma_half_time()
    price_scale = vyper_cryptopool.price_scale()
    price_oracle = vyper_cryptopool.eval("self._price_oracle")
    last_prices = vyper_cryptopool.last_prices()
    last_prices_timestamp = vyper_cryptopool.last_prices_timestamp()
    balances = [vyper_cryptopool.balances(i) for i in range(n_coins)]
    # Use the cached `D`. See warning for `virtual_price` below
    D = vyper_cryptopool.D()
    lp_total_supply = vyper_cryptopool.totalSupply()
    xcp_profit = vyper_cryptopool.xcp_profit()
    xcp_profit_a = vyper_cryptopool.xcp_profit_a()

    pool = CurveCryptoPool(
        A=A,
        gamma=gamma,
        n=n_coins,
        precisions=precisions,
        mid_fee=mid_fee,
        out_fee=out_fee,
        allowed_extra_profit=allowed_extra_profit,
        fee_gamma=fee_gamma,
        adjustment_step=adjustment_step,
        admin_fee=admin_fee,
        ma_half_time=ma_half_time,
        price_scale=[price_scale],
        price_oracle=[price_oracle],
        last_prices=[last_prices],
        last_prices_timestamp=last_prices_timestamp,
        balances=balances,
        D=D,
        tokens=lp_total_supply,
        xcp_profit=xcp_profit,
        xcp_profit_a=xcp_profit_a,
    )

    assert pool.A == vyper_cryptopool.A()
    assert pool.gamma == vyper_cryptopool.gamma()
    assert pool.balances == balances
    assert pool.price_scale == [price_scale]
    assert pool._price_oracle == [price_oracle]  # pylint: disable=protected-access
    assert pool.last_prices == [last_prices]
    assert pool.last_prices_timestamp == last_prices_timestamp
    assert pool.D == vyper_cryptopool.D()
    assert pool.tokens == lp_total_supply
    assert pool.xcp_profit == xcp_profit
    assert pool.xcp_profit_a == xcp_profit_a

    # WARNING: both `virtual_price` and `D` are cached values
    # so depending on the test, may not be updated to be
    # consistent with the rest of the pool state.
    #
    # Allowing this simplifies testing since we can avoid
    # coupling tests of basic functionality with the tests
    # for the complex newton calculations.
    #
    # We think it makes sense the initialized pool should
    # at least match the vyper pool (inconsistencies and all).
    # When full consistency is required, the `update_cached_values`
    # helper function should be utilized before calling
    # `initialize_pool`.
    virtual_price = vyper_cryptopool.virtual_price()
    pool.virtual_price = virtual_price

    return pool


def sync_ema_logic(
    vyper_cryptopool,
    pool,
    last_prices,
):
    """
    Test helper to synchronize state variables needed for EMA update.

    This is needed because the local evm block timestamp will drift
    from the python pool's internal timestamp.
    """
    # pylint: disable=protected-access
    price_oracle = vyper_cryptopool.eval("self._price_oracle")
    pool._price_oracle = [price_oracle]

    # synchronize the times between the two pools and reset
    # last_prices and last_prices_timestamp
    vyper_cryptopool.eval(f"self.last_prices={last_prices}")
    pool.last_prices = [last_prices]

    vm_timestamp = boa.env.vm.state.timestamp
    pool._increment_timestamp(timestamp=vm_timestamp)

    last_prices_timestamp = vm_timestamp - 120
    vyper_cryptopool.eval(f"self.last_prices_timestamp={last_prices_timestamp}")
    pool.last_prices_timestamp = last_prices_timestamp


def pack_A_gamma(A, gamma):
    """
    Need this to set A and gamma in the smart contract since they
    are stored in packed format.
    """
    A_gamma = A << 128
    A_gamma = A_gamma | gamma
    return A_gamma


def get_real_balances(virtual_balances, precisions, price_scale):
    """
    Convert from units of D to native token units using the
    given price scale.
    """
    assert len(virtual_balances) == 2
    balances = [x // p for x, p in zip(virtual_balances, precisions)]
    balances[1] = balances[1] * PRECISION // price_scale
    return balances


def update_cached_values(vyper_cryptopool):
    """
    Useful test helper after we manipulate the pool state.

    Calculates `D` and `virtual_price` from pool state and caches
    them in the appropriate storage.
    """
    A = vyper_cryptopool.A()
    gamma = vyper_cryptopool.gamma()
    xp = vyper_cryptopool.eval("self.xp()")
    xp = list(xp)  # boa doesn't like its own tuple wrapper
    D = vyper_cryptopool.eval(f"self.newton_D({A}, {gamma}, {xp})")
    vyper_cryptopool.eval(f"self.D={D}")
    total_supply = vyper_cryptopool.totalSupply()
    vyper_cryptopool.eval(
        f"self.virtual_price=10**18 * self.get_xcp({D})/{total_supply}"
    )


D_UNIT = 10**18
positive_balance = st.integers(min_value=10**5 * D_UNIT, max_value=10**11 * D_UNIT)
amplification_coefficient = st.integers(min_value=MIN_A, max_value=MAX_A)
gamma_coefficient = st.integers(min_value=MIN_GAMMA, max_value=MAX_GAMMA)
price = st.integers(min_value=10**12, max_value=10**25)


@given(positive_balance, positive_balance)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=2,
    deadline=None,
)
def test_xp(vyper_cryptopool, x0, x1):
    """Test xp calculation against vyper implementation."""

    _balances = [x0, x1]
    precisions = vyper_cryptopool.eval("self._get_precisions()")
    price_scale = vyper_cryptopool.price_scale()
    balances = get_real_balances(_balances, precisions, price_scale)

    vyper_cryptopool.eval(f"self.balances={balances}")
    expected_xp = vyper_cryptopool.eval("self.xp()")
    expected_xp = list(expected_xp)

    pool = initialize_pool(vyper_cryptopool)
    xp = pool._xp()  # pylint: disable=protected-access

    assert xp == expected_xp


@given(positive_balance, positive_balance, st.booleans())
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=2,
    deadline=None,
)
def test_geometric_mean(vyper_cryptopool, x0, x1, sort_flag):
    """Test geometric_mean calculation against vyper implementation."""

    xp = [x0, x1]
    expected_result = vyper_cryptopool.eval(f"self.geometric_mean({xp}, {sort_flag})")
    result = geometric_mean(xp, sort_flag)

    assert result == expected_result


@given(st.integers(min_value=0))
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=2,
    deadline=None,
)
def test_halfpow(vyper_cryptopool, power):
    """Test halfpow calculation against vyper implementation."""

    expected_result = vyper_cryptopool.eval(f"self.halfpow({power})")
    result = halfpow(power)

    assert result == expected_result


@given(st.integers(min_value=0))
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=2,
    deadline=None,
)
def test_sqrt_int(vyper_cryptopool, number):
    """Test sqrt_int calculation against vyper implementation."""

    expected_result = vyper_cryptopool.eval(f"self.sqrt_int({number})")
    result = _sqrt_int(number)  # pylint: disable=protected-access

    assert result == expected_result


@given(st.integers(min_value=100))
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=2,
    deadline=None,
)
def test_get_xcp(vyper_cryptopool, D):
    """Test get_xcp calculation against vyper implementation."""

    expected_result = vyper_cryptopool.eval(f"self.get_xcp({D})")

    pool = initialize_pool(vyper_cryptopool)
    result = pool._get_xcp(D)  # pylint: disable=protected-access

    assert result == expected_result


@given(amplification_coefficient, gamma_coefficient, positive_balance, positive_balance)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_newton_D(vyper_cryptopool, A, gamma, x0, x1):
    """Test D calculation against vyper implementation."""

    xp = [x0, x1]
    assume(0.02 < xp[0] / xp[1] < 50)

    expected_D = vyper_cryptopool.eval(f"self.newton_D({A}, {gamma}, {xp})")
    D = newton_D(A, gamma, xp)

    assert D == expected_D


@given(
    amplification_coefficient,
    gamma_coefficient,
    positive_balance,
    positive_balance,
    st.integers(min_value=0, max_value=1),
    st.integers(min_value=0, max_value=25),
)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_get_y(vyper_cryptopool, A, gamma, x0, x1, i, delta_perc):
    """Test get_y calculation against vyper implementation."""

    xp = [x0, x1]
    assume(0.02 < xp[0] / xp[1] < 50)

    # vary D by delta_perc %
    D = vyper_cryptopool.eval(f"self.newton_D({A}, {gamma}, {xp})")
    D_changed = D * (100 - delta_perc) // 100
    expected_y = vyper_cryptopool.eval(
        f"self.newton_y({A}, {gamma}, {xp}, {D_changed}, {i})"
    )

    y, _ = get_y(A, gamma, xp, D_changed, i)

    assert y == expected_y

    # vary xp[j] by delta_perc %
    xp_changed = xp.copy()
    j = 1 - i
    xp_changed[j] = xp[j] * (100 - delta_perc) // 100
    expected_y = vyper_cryptopool.eval(
        f"self.newton_y({A}, {gamma}, {xp_changed}, {D}, {i})"
    )

    y, _ = get_y(A, gamma, xp_changed, D, i)

    assert y == expected_y


@given(
    amplification_coefficient,
    gamma_coefficient,
    st.integers(min_value=10**6 * D_UNIT, max_value=10**9 * D_UNIT),
    st.integers(min_value=10, max_value=1000),
    st.integers(min_value=70, max_value=200).filter(lambda x: x < 90 or x > 110),
)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_tweak_price(
    vyper_cryptopool, cryptopool_lp_token, A, gamma, x0, x_perc, price_perc
):
    # def test_tweak_price(vyper_cryptopool, cryptopool_lp_token):
    #     A = 4196
    #     gamma = 10000050055
    #     x0 = 100000000000000000017267
    #     x1 = 100000000000000000000153
    #     last_price = 1000000000219

    # pylint: disable=protected-access
    x1 = x0 * x_perc // 100
    _balances = [x0, x1]
    price_oracle = vyper_cryptopool.eval("self._price_oracle")
    last_price = price_oracle * price_perc // 100

    precisions = vyper_cryptopool.eval("self._get_precisions()")
    price_scale = vyper_cryptopool.price_scale()
    balances = get_real_balances(_balances, precisions, price_scale)

    vyper_cryptopool.eval(f"self.balances={balances}")
    xp = vyper_cryptopool.eval("self.xp()")
    xp = list(xp)

    A_gamma = [A, gamma]
    # need to set A_gamma since this is read when claiming admin fees
    A_gamma_packed = pack_A_gamma(A, gamma)
    vyper_cryptopool.eval(f"self.future_A_gamma={A_gamma_packed}")

    # need to update cached `D` and `virtual_price`
    # (this requires adjusting LP token supply for consistency)
    D = vyper_cryptopool.eval(f"self.newton_D({A}, {gamma}, {xp})")
    vyper_cryptopool.eval(f"self.D={D}")

    totalSupply = vyper_cryptopool.eval(f"self.get_xcp({D})")
    cryptopool_lp_token.eval(f"self.totalSupply={totalSupply}")
    # virtual price can't be below 10**18
    vyper_cryptopool.eval("self.virtual_price=10**18")
    # reset profit counter also
    vyper_cryptopool.eval("self.xcp_profit=10**18")
    vyper_cryptopool.eval("self.xcp_profit_a=10**18")

    pool = initialize_pool(vyper_cryptopool)

    # ------- test no oracle update and no scale adjustment ------------- #
    assert pool.price_scale == [vyper_cryptopool.price_scale()]
    assert pool._price_oracle == [vyper_cryptopool.eval("self._price_oracle")]

    old_scale = pool.price_scale
    assert old_scale == pool._price_oracle

    old_oracle = pool._price_oracle

    # pylint: disable=protected-access
    pool._tweak_price(A, gamma, xp, 1, last_price, 0)
    vyper_cryptopool.eval(f"self.tweak_price({A_gamma}, {xp}, {last_price}, 0)")

    assert pool.price_scale == [vyper_cryptopool.price_scale()]
    # no price adjustment since price oracle is same as price scale (`norm` is 0)
    assert pool.price_scale == old_scale
    # EMA price oracle won't update if price oracle and last price is the same
    assert old_oracle == pool._price_oracle

    # ------- test oracle updates with no scale adjustment ------------- #

    sync_ema_logic(vyper_cryptopool, pool, last_price)

    assert pool.virtual_price == vyper_cryptopool.virtual_price()
    assert pool.xcp_profit == vyper_cryptopool.xcp_profit()

    old_oracle = pool._price_oracle
    old_scale = pool.price_scale
    old_virtual_price = pool.virtual_price

    pool._tweak_price(A, gamma, xp, 1, last_price, 0)
    vyper_cryptopool.eval(f"self.tweak_price({A_gamma}, {xp}, {last_price}, 0)")

    # check the pools are the same
    assert pool.price_scale == [vyper_cryptopool.price_scale()]
    assert pool._price_oracle == [vyper_cryptopool.eval("self._price_oracle")]
    assert pool.virtual_price == vyper_cryptopool.virtual_price()
    assert pool.D == vyper_cryptopool.D()

    # check oracle updated
    # scale shouldn't change because no adjustment is possible
    # with no profit
    assert pool._price_oracle != old_oracle
    assert pool.price_scale == old_scale
    assert pool.virtual_price == old_virtual_price

    # ------- test scale adjustment ----------------------- #
    sync_ema_logic(vyper_cryptopool, pool, last_price)

    assert pool.price_scale != pool._price_oracle

    old_oracle = pool._price_oracle
    old_scale = pool.price_scale
    old_virtual_price = pool.virtual_price

    xp[0] = xp[0] + pool.allowed_extra_profit // 10

    # omitting price will calculate the spot price in `tweak_price`
    pool._tweak_price(A, gamma, xp, 1, 0, 0)
    vyper_cryptopool.eval(f"self.tweak_price({A_gamma}, {xp}, 0, 0)")

    assert pool.price_scale == [vyper_cryptopool.price_scale()]
    assert pool._price_oracle == [vyper_cryptopool.eval("self._price_oracle")]

    assert pool.D == vyper_cryptopool.D()
    assert pool.virtual_price == vyper_cryptopool.virtual_price()
    assert pool.xcp_profit == vyper_cryptopool.xcp_profit()

    # profit increased but not enough to adjust the price scale
    assert pool.virtual_price > old_virtual_price
    assert pool.price_scale == old_scale

    old_virtual_price = pool.virtual_price

    xp[0] = xp[0] * 115 // 100

    # omitting price will calculate the spot price in `tweak_price`
    pool._tweak_price(A, gamma, xp, 1, 0, 0)
    vyper_cryptopool.eval(f"self.tweak_price({A_gamma}, {xp}, 0, 0)")

    assert pool.price_scale == [vyper_cryptopool.price_scale()]
    assert pool._price_oracle == [vyper_cryptopool.eval("self._price_oracle")]

    assert pool.D == vyper_cryptopool.D()
    assert pool.virtual_price == vyper_cryptopool.virtual_price()
    assert pool.xcp_profit == vyper_cryptopool.xcp_profit()

    # profit increased enough to adjust the price scale
    assert pool.virtual_price > old_virtual_price
    assert pool.price_scale != old_scale


@given(
    positive_balance,
    positive_balance,
    st.integers(min_value=1, max_value=10**4),
    st.integers(min_value=0, max_value=1),
)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_get_dy(vyper_cryptopool, x0, x1, dx_perc, i):
    """Test get_dy calculation against vyper implementation."""
    assume(0.02 < x0 / x1 < 50)

    j = 1 - i
    xp = [x0, x1]

    precisions = vyper_cryptopool.eval("self._get_precisions()")
    price_scale = vyper_cryptopool.price_scale()
    balances = get_real_balances(xp, precisions, price_scale)

    vyper_cryptopool.eval(f"self.balances={balances}")
    update_cached_values(vyper_cryptopool)
    pool = initialize_pool(vyper_cryptopool)

    dx = balances[i] * dx_perc // 10**5

    expected_dy = vyper_cryptopool.get_dy(i, j, dx)
    dy = pool.get_dy(i, j, dx)

    assert dy == expected_dy

    expected_balances = [vyper_cryptopool.balances(i) for i in range(2)]
    assert pool.balances == expected_balances


@given(
    st.integers(min_value=1, max_value=300),
    st.integers(min_value=0, max_value=1),
)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_exchange(vyper_cryptopool, dx_perc, i):
    """Test `exchange` against vyper implementation."""
    j = 1 - i

    pool = initialize_pool(vyper_cryptopool)
    dx = pool.balances[i] * dx_perc // 100

    expected_dy = vyper_cryptopool.exchange(i, j, dx, 0)
    dy, _ = pool.exchange(i, j, dx)

    assert dy == expected_dy

    expected_balances = [vyper_cryptopool.balances(i) for i in range(2)]
    assert pool.balances == expected_balances


@given(positive_balance, positive_balance)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_add_liquidity(vyper_cryptopool, x0, x1):
    """Test `add_liquidity` against vyper implementation."""
    assume(0.02 < x0 / x1 < 50)
    xp = [x0, x1]

    precisions = vyper_cryptopool.eval("self._get_precisions()")
    price_scale = vyper_cryptopool.price_scale()
    amounts = get_real_balances(xp, precisions, price_scale)

    pool = initialize_pool(vyper_cryptopool)

    expected_lp_amount = vyper_cryptopool.add_liquidity(amounts, 0)
    expected_balances = [vyper_cryptopool.balances(i) for i in range(len(xp))]
    expected_lp_supply = vyper_cryptopool.totalSupply()
    expected_D = vyper_cryptopool.D()

    lp_amount = pool.add_liquidity(amounts)

    assert lp_amount == expected_lp_amount
    assert pool.balances == expected_balances
    assert pool.tokens == expected_lp_supply
    assert pool.D == expected_D


@given(positive_balance)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_remove_liquidity(vyper_cryptopool, amount):
    """Test `remove_liquidity` against vyper implementation."""
    assume(amount < vyper_cryptopool.totalSupply())

    pool = initialize_pool(vyper_cryptopool)

    vyper_cryptopool.remove_liquidity(amount, [0, 0])
    expected_balances = [vyper_cryptopool.balances(i) for i in range(2)]
    expected_lp_supply = vyper_cryptopool.totalSupply()
    expected_D = vyper_cryptopool.D()

    pool.remove_liquidity(amount)

    assert pool.balances == expected_balances
    assert pool.tokens == expected_lp_supply
    assert pool.D == expected_D


@given(positive_balance, st.integers(min_value=0, max_value=1))
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_remove_liquidity_one_coin(vyper_cryptopool, amount, i):
    """Test `remove_liquidity_one_coin` against vyper implementation."""
    assume(amount < vyper_cryptopool.totalSupply())

    pool = initialize_pool(vyper_cryptopool)

    vyper_cryptopool.remove_liquidity_one_coin(amount, i, 0)
    expected_coin_balance = vyper_cryptopool.balances(i)
    expected_lp_supply = vyper_cryptopool.totalSupply()

    pool.remove_liquidity_one_coin(amount, i, 0)
    coin_balance = pool.balances[i]
    lp_supply = pool.tokens

    assert coin_balance == expected_coin_balance
    assert lp_supply == expected_lp_supply


@given(positive_balance, st.integers(min_value=0, max_value=1))
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=25,
    deadline=None,
)
def test_calc_withdraw_one_coin(vyper_cryptopool, amount, i):
    """Test `calc_withdraw_one_coin` against vyper implementation."""
    assume(amount < vyper_cryptopool.totalSupply())

    pool = initialize_pool(vyper_cryptopool)

    expected_dy = vyper_cryptopool.calc_withdraw_one_coin(amount, i)
    dy = pool.calc_withdraw_one_coin(amount, i)
    assert dy == expected_dy

    expected_balances = [vyper_cryptopool.balances(i) for i in range(2)]
    assert pool.balances == expected_balances


@given(price, price, st.integers(min_value=0, max_value=1000))
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_price_oracle(vyper_cryptopool, price_oracle, last_price, time_delta):
    """Test `price_oracle` and `lp_price` against vyper implementation."""
    vyper_cryptopool.eval(f"self._price_oracle={price_oracle}")
    vyper_cryptopool.eval(f"self.last_prices={last_price}")
    vm_timestamp = boa.env.vm.state.timestamp
    last_prices_timestamp = vm_timestamp - time_delta
    vyper_cryptopool.eval(f"self.last_prices_timestamp={last_prices_timestamp}")

    pool = initialize_pool(vyper_cryptopool)
    # pylint: disable-next=protected-access
    pool._increment_timestamp(timestamp=vm_timestamp)

    assert pool.price_oracle() == [vyper_cryptopool.price_oracle()]
    assert pool.lp_price() == vyper_cryptopool.lp_price()


@given(positive_balance, positive_balance, price)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_get_virtual_price(
    vyper_cryptopool, cryptopool_lp_token, D, tokens, price_scale
):
    vyper_cryptopool.eval(f"self.D={D}")
    cryptopool_lp_token.eval(f"self.totalSupply={tokens}")
    vyper_cryptopool.eval(f"self.price_scale={price_scale}")
    pool = initialize_pool(vyper_cryptopool)

    expected_virtual_price = vyper_cryptopool.get_virtual_price()
    virtual_price = pool.get_virtual_price()
    assert virtual_price == expected_virtual_price


@given(positive_balance, positive_balance)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_calc_token_amount(vyper_cryptopool, x0, x1):
    """Test `calc_token_amount` against vyper implementation."""
    assume(0.02 < x0 / x1 < 50)
    xp = [x0, x1]

    precisions = vyper_cryptopool.eval("self._get_precisions()")
    price_scale = vyper_cryptopool.price_scale()
    amounts = get_real_balances(xp, precisions, price_scale)

    pool = initialize_pool(vyper_cryptopool)

    expected_lp_amount = vyper_cryptopool.calc_token_amount(amounts)
    lp_amount = pool.calc_token_amount(amounts)
    assert lp_amount == expected_lp_amount

    expected_balances = [vyper_cryptopool.balances(i) for i in range(2)]
    assert pool.balances == expected_balances


_num_iter = 10


@given(
    st.lists(
        st.integers(min_value=1, max_value=5000),
        min_size=_num_iter,
        max_size=_num_iter,
    ),
    st.lists(
        st.tuples(
            st.integers(min_value=0, max_value=1),
            st.integers(min_value=0, max_value=1),
        ).filter(lambda x: x[0] != x[1]),
        min_size=_num_iter,
        max_size=_num_iter,
    ),
    st.lists(
        st.integers(min_value=0, max_value=86400),
        min_size=_num_iter,
        max_size=_num_iter,
    ),
)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_multiple_exchange_with_repeg(
    vyper_cryptopool, dx_perc_list, indices_list, time_delta_list
):
    """Test `exchange` against vyper implementation."""

    pool = initialize_pool(vyper_cryptopool)

    for indices, dx_perc, time_delta in zip(
        indices_list, dx_perc_list, time_delta_list
    ):
        vm_timestamp = boa.env.vm.state.timestamp
        pool._block_timestamp = vm_timestamp

        i, j = indices
        dx = pool.balances[i] * dx_perc // 10000  # dx_perc in bps

        expected_dy = vyper_cryptopool.exchange(i, j, dx, 0)
        dy, _ = pool.exchange(i, j, dx)
        assert dy == expected_dy

        expected_balances = [vyper_cryptopool.balances(i) for i in range(2)]
        assert pool.balances[0] == expected_balances[0]
        assert pool.balances[1] == expected_balances[1]

        assert pool.last_prices == [vyper_cryptopool.last_prices()]
        assert pool.last_prices_timestamp == vyper_cryptopool.last_prices_timestamp()

        expected_price_oracle = [vyper_cryptopool.price_oracle()]
        expected_price_scale = [vyper_cryptopool.price_scale()]
        assert pool.price_oracle() == expected_price_oracle
        assert pool.price_scale == expected_price_scale

        boa.env.time_travel(time_delta)


def test_dydxfee(vyper_cryptopool):
    """Test spot price formula against execution price for small trades."""
    pool = initialize_pool(vyper_cryptopool)

    # STG, USDC
    decimals = [18, 6]
    precisions = [10 ** (18 - d) for d in decimals]

    i = 0
    j = 1
    dx = 10**18

    dydx = pool.dydxfee(i, j)
    dy = vyper_cryptopool.exchange(i, j, dx, 0)

    dx *= precisions[i]
    dy *= precisions[j]
    assert abs(dydx - dy / dx) < 1e-6



"""Unit tests for the SnapshotMixin and derived subclasses of Snapshot."""
import pytest

from curvesim.exceptions import SnapshotError
from curvesim.pool.sim_interface import SimCurveMetaPool, SimCurvePool
from curvesim.pool.snapshot import SnapshotMixin


def test_snapshot_raises_exception():
    """Test error thrown for missing snapshot class."""
    obj = SnapshotMixin()
    with pytest.raises(SnapshotError):
        obj.get_snapshot()


def test_pool_balance_snapshot():
    """Test balances are copied and reset properly for SimCurvePool."""
    pool = SimCurvePool(A=250, D=1000000 * 10**18, n=2, admin_fee=5 * 10**9)

    snapshot = pool.get_snapshot()

    pre_balances = pool.balances.copy()
    pre_admin_balances = pool.admin_balances.copy()
    pool.exchange(0, 1, 10**12)
    post_balances = pool.balances.copy()
    post_admin_balances = pool.admin_balances.copy()
    assert pre_balances != post_balances
    assert pre_admin_balances != post_admin_balances

    pool.revert_to_snapshot(snapshot)
    assert pool.balances == pre_balances
    assert pool.admin_balances == pre_admin_balances


def test_metapool_balance_snapshot():
    """Test balances are copied and reset properly for SimCurveMetaPool."""
    basepool = SimCurvePool(A=1000, D=2750000 * 10**18, n=2, admin_fee=5 * 10**9)
    pool = SimCurveMetaPool(
        A=250, D=4000000 * 10**18, n=2, admin_fee=5 * 10**9, basepool=basepool
    )

    snapshot = pool.get_snapshot()

    pre_balances = pool.balances.copy()
    pre_bp_balances = basepool.balances.copy()
    pre_admin_balances = pool.admin_balances.copy()
    pre_bp_admin_balances = basepool.admin_balances.copy()
    pre_lp_tokens = basepool.tokens

    # ---- test top-level exchange ---- #

    pool.exchange(0, 1, 10**12)

    # `exchange` changes values for metapool not basepool
    assert pool.balances != pre_balances
    assert pool.admin_balances != pre_admin_balances
    assert basepool.balances == pre_bp_balances
    assert basepool.admin_balances == pre_bp_admin_balances
    assert basepool.tokens == pre_lp_tokens

    pool.revert_to_snapshot(snapshot)

    assert pool.balances == pre_balances
    assert pool.admin_balances == pre_admin_balances
    assert basepool.balances == pre_bp_balances
    assert basepool.admin_balances == pre_bp_admin_balances

    # -- test exchange between primary and basepool underlyer -- #

    pool.exchange_underlying(0, 1, 157 * 10**18)

    # `exchange_underlying` should change basepool values
    assert pool.balances != pre_balances
    assert pool.admin_balances != pre_admin_balances
    assert basepool.balances != pre_bp_balances
    assert basepool.admin_balances != pre_bp_admin_balances
    assert basepool.tokens != pre_lp_tokens

    pool.revert_to_snapshot(snapshot)

    assert pool.balances == pre_balances
    assert pool.admin_balances == pre_admin_balances
    assert basepool.balances == pre_bp_balances
    assert basepool.admin_balances == pre_bp_admin_balances
    assert basepool.tokens == pre_lp_tokens


def test_snapshot_context():
    """Test context manager."""
    pool = SimCurvePool(A=250, D=1000000 * 10**18, n=2, admin_fee=5 * 10**9)

    pre_balances = pool.balances.copy()
    pre_admin_balances = pool.admin_balances.copy()

    with pool.use_snapshot_context():
        pool.exchange(0, 1, 10**12)
        assert pool.balances != pre_balances
        assert pool.admin_balances != pre_admin_balances

    assert pool.balances == pre_balances
    assert pool.admin_balances == pre_admin_balances



"""
This module provides functionality for creating result selectors for an Altair
chart.

It contains utility functions for creating axis selectors, parameter filters,
selector charts and result selectors.
"""

from altair import CalculateTransform, Color, FilterTransform, Scale, concat, hconcat

from ..selectors import make_selector


def make_result_selectors(factors, dynamic_axes):
    """
    Creates result selectors for an Altair chart.

    Parameters
    ----------
    factors : dict
        The factors to create selectors for.
    dynamic_axes : dict
        The dynamic axes to create selectors for.

    Returns
    -------
    dict
        Keyword arguments to add selections to metric charts and chart objects to
        display the selectors.
    """
    dynamic_axes = dict(list(dynamic_axes.items())[: len(factors)])

    axes = make_axis_selectors(factors, dynamic_axes)
    parameters = make_parameter_filters(factors, dynamic_axes)

    # kwargs to pass to metric charts
    metric_kwargs = {
        axis: {"shorthand": shorthand, "title": None}
        for axis, shorthand in dynamic_axes.items()
    }
    metric_kwargs["transform"] = axes["transforms"] + parameters["transforms"]

    return {
        "kwargs": metric_kwargs,
        "charts": format_selector_charts(axes["charts"], parameters["charts"]),
    }


def make_axis_selectors(factors, dynamic_axes):
    """
    Creates axis selectors.

    Parameters
    ----------
    factors : dict
        The factors to use as selector options.
    dynamic_axes : dict
        The dynamic axes to create selectors for.

    Returns
    -------
    dict
        The axis selectors and the charts to display them.
    """
    factor_names = list(factors.keys())

    charts = []
    transforms = []
    for idx, axis in enumerate(dynamic_axes):
        chart, transform = _make_axis_selector(axis, factor_names, idx)

        charts.append(chart)
        transforms.append(transform)

    return {"charts": charts, "transforms": transforms}


def make_parameter_filters(factors, dynamic_axes):
    """
    Creates parameter filters.

    Parameters
    ----------
    factors : dict
        The factors to create selectors for.
    dynamic_axes : dict
        The dynamic axes for a chart page. Used to determine initial selections.

    Returns
    -------
    dict
        The parameter filters and the charts to display them.
    """
    n_dynamic_axes = len(dynamic_axes)

    charts = []
    transforms = []
    for i, factor in enumerate(factors.items()):
        idx = "all" if i < n_dynamic_axes else 0
        chart, transform = _make_parameter_filter(*factor, idx)

        charts.append(chart)
        transforms.append(transform)

    return {"charts": charts, "transforms": transforms}


def _make_axis_selector(axis, options, sel_idx):
    """
    Makes a single axis selector.

    Parameters
    ----------
    axis : str
        The axis to create a selector for.
    options : list
        The options for the selector.
    sel_idx : int or 'all'
        The initial selection index. If an integer, select that option. If 'all',
        select all options.

    Returns
    -------
    altair.Chart, altair.CalculateTransform
        The axis selector chart and the transform for the selector.
    """
    chart, selector = make_selector(
        axis,
        options,
        title=axis,
        style={"axis": "y", "orient": "left"},
        toggle=False,
        sel_idx=sel_idx,
    )

    calc_kwargs = {"calculate": f"datum[{selector.name}.{axis}]", "as": axis}
    return chart, CalculateTransform(**calc_kwargs)


def _make_parameter_filter(factor, options, sel_idx):
    """
    Makes a single parameter filter.

    Parameters
    ----------
    factor : str
        The factor to create a filter for.
    options : list
        The options for the filter.
    sel_idx : int or 'all'
        The initial selection index. If an integer, select that option. If 'all',
        select all options.

    Returns
    -------
    altair.Chart, altair.FilterTransform
        The parameter filter chart and the filter transform.
    """
    color = Color("labels:O", scale=Scale(scheme="viridis"), legend=None)

    chart, selector = make_selector(
        factor,
        options,
        title=factor,
        style={"axis": "x", "orient": "bottom", "color": color},
        toggle="true",
        sel_idx=sel_idx,
    )
    return chart, FilterTransform(selector)


def format_selector_charts(axis_selector_charts, parameter_filter_charts):
    """
    Formats selector charts for display.

    Parameters
    ----------
    axis_selector_charts : list of altair.Chart
        The axis selector charts.
    parameter_filter_charts : list of altair.Chart
        The parameter filter charts.

    Returns
    -------
    altair.hconcat
        The formatted selector charts.
    """
    left = concat(*axis_selector_charts, title="Axis Selectors:")
    right = concat(*parameter_filter_charts, title="Toggle Filters:")
    return hconcat(left, right.resolve_scale(color="independent"))



from abc import ABC, abstractmethod
from typing import Union

from curvesim.logging import get_logger
from curvesim.utils import dataclass

logger = get_logger(__name__)


@dataclass(frozen=True, slots=True)
class Trade:
    """Data container for a trade to be executed."""

    coin_in: Union[str, int]
    coin_out: Union[str, int]
    amount_in: int

    def __iter__(self):
        # pylint: disable=no-member
        return (getattr(self, attr) for attr in self.__slots__)


@dataclass(slots=True)
class TradeResult:
    """Data container for a trade execution."""

    coin_in: Union[str, int]
    coin_out: Union[str, int]
    amount_in: int
    amount_out: int
    fee: int

    def __iter__(self):
        # pylint: disable=no-member
        return (getattr(self, attr) for attr in self.__slots__)

    def set_attrs(self, **kwargs):
        """Sets multiple attributes defined by keyword arguments."""
        for attr, value in kwargs.items():
            setattr(self, attr, value)

    @classmethod
    def from_trade(cls, trade, amount_out=None, fee=None):
        """Initializes a TradeResult object from a Trade object"""
        return cls(trade.coin_in, trade.coin_out, trade.amount_in, amount_out, fee)


class Trader(ABC):
    """
    Computes, executes, and reports out arbitrage trades.
    """

    def __init__(self, pool):
        """
        Parameters
        ----------
        pool :
            Simulation interface to a subclass of :class:`.Pool`.

        """
        self.pool = pool

    @abstractmethod
    def compute_trades(self, *args):
        """
        Computes trades to execute on the pool.

        Returns
        -------
        trades : list of :class:`Trade` objects
            List of trades to perform.

        additional_data: dict
            Dict of additional data to be passed to the state log as part of trade_data.
        """
        raise NotImplementedError

    def do_trades(self, trades):
        """
        Executes a series of trades.

        Parameters
        ----------
        trades : list of :class:`Trade` objects
            Trades to execute.

        Returns
        -------
        trades: list of :class:`TradeResult` objects
            The results of the trades.

        """

        trade_results = []
        for trade in trades:
            dy, fee = self.pool.trade(trade.coin_in, trade.coin_out, trade.amount_in)
            trade_results.append(TradeResult.from_trade(trade, amount_out=dy, fee=fee))

        return trade_results

    def process_time_sample(self, *args):
        """
        Process given tick data by computing and executing trades.

        The input args must be properly formed and fed by the
        parent `Strategy` object housing the trader class via its
        :meth:`~curvesim.pipelines.templates.Strategy._get_trader_inputs`.
        """
        trades, additional_data = self.compute_trades(*args)
        trade_results = self.do_trades(trades)

        return {"trades": trade_results, **additional_data}



"""
Helper functions for the different data sources we pull from.
"""
from datetime import datetime, timezone

from curvesim.logging import get_logger
from curvesim.network import coingecko as _coingecko
from curvesim.network import nomics as _nomics

logger = get_logger(__name__)


def coingecko(coins, chain="mainnet", days=60, end=None):
    """
    Fetch CoinGecko price data for specified coins.

    Parameters
    ----------
    coins : list of str
        List of coin symbols to fetch data for.
    chain : str, optional
        Blockchain network to consider. Default is "mainnet".
    days : int, optional
        Number of past days to fetch data for. Default is 60.
    end : int, optional
        End timestamp for the data in seconds since epoch.
        If None, the end time will be the current time. Default is None.

    Returns
    -------
    tuple of (dict, dict, int)
        Tuple of prices, volumes, and pzero (fixed as 0 for this function).
    """
    logger.info("Fetching CoinGecko price data...")
    prices, volumes = _coingecko.pool_prices(coins, "usd", days, chain=chain, end=end)
    pzero = 0

    return prices, volumes, pzero


def local(coins, data_dir="data", end=None):
    """
    Load data for specified coins from a local directory.

    Parameters
    ----------
    coins : list of str
        List of coin symbols to load data for.
    data_dir : str, optional
        Path to the directory containing the data files. Default is "data".
    end : int, optional
        End timestamp for the data in seconds since epoch.
        If None, the end time will be the current time. Default is None.

    Returns
    -------
    tuple of (dict, dict, int)
        Tuple of prices, volumes, and pzero.
    """
    logger.info("Using local data...")
    if end is not None:
        t_end = datetime.fromtimestamp(end, tz=timezone.utc)
        custom_suffix = "-" + str(end)
    else:
        t_end = None
        custom_suffix = ""
    prices, volumes, pzero = _nomics.local_pool_prices(
        coins, data_dir=data_dir, t_end=t_end, custom_suffix=custom_suffix
    )

    return prices, volumes, pzero



"""Unit tests for CurveMetaPool"""
import pytest
from hypothesis import HealthCheck, assume, given, settings
from hypothesis import strategies as st

from curvesim.pool import CurveMetaPool, CurvePool

from .test_pool import initialize_pool


def convert_to_virtual_balances(rates, balances):
    """Convert from native token units to D units."""
    return [b * p // 10**18 for b, p in zip(balances, rates)]


def convert_to_real_balances(rates, balances):
    """Convert from D units to native token units."""
    return [b * 10**18 // p for b, p in zip(balances, rates)]


def initialize_metapool(vyper_metapool, vyper_3pool):
    """
    Initialize python-based pool from the state variables of the
    vyper-based implementation.
    """
    A = vyper_metapool.A()
    n_coins = vyper_metapool.N_COINS()
    balances = [vyper_metapool.balances(i) for i in range(n_coins)]
    lp_total_supply = vyper_metapool.totalSupply()
    fee = vyper_metapool.fee()
    admin_fee = vyper_metapool.admin_fee()
    rate_multiplier = vyper_metapool.rate_multiplier()
    basepool = initialize_pool(vyper_3pool)
    metapool = CurveMetaPool(
        A,
        D=balances,
        n=n_coins,
        basepool=basepool,
        rate_multiplier=rate_multiplier,
        tokens=lp_total_supply,
        fee=fee,
        admin_fee=admin_fee,
    )
    return metapool


# We can assume the contract works on more extreme values; we only need
# to be reasonably certain our results are consistent, so we can check
# a smaller range.
#
# With 18 decimal precision, it seems reasonable to pick these bounds
D_UNIT = 10**18
positive_balance = st.integers(min_value=10**5 * D_UNIT, max_value=10**10 * D_UNIT)


@given(positive_balance, positive_balance)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_get_D(vyper_metapool, vyper_3pool, x0, x1):
    """Test D calculation against vyper implementation."""

    _balances = [x0, x1]
    p = [vyper_metapool.rates(i) for i in range(len(_balances))]
    balances = convert_to_real_balances(p, _balances)

    vyper_metapool.eval(f"self.balances={balances}")
    expected_D = vyper_metapool.D()

    python_metapool = initialize_metapool(vyper_metapool, vyper_3pool)
    D = python_metapool.D()

    assert D == expected_D


def test_get_virtual_price(vyper_metapool, vyper_3pool):
    """Test `get_virtual_price` against vyper implementation."""
    python_metapool = initialize_metapool(vyper_metapool, vyper_3pool)
    virtual_price = python_metapool.get_virtual_price()
    expected_virtual_price = vyper_metapool.get_virtual_price()
    assert virtual_price == expected_virtual_price


@given(
    positive_balance,
    st.integers(min_value=0, max_value=1),
    st.integers(min_value=0, max_value=1),
)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_get_y(vyper_metapool, vyper_3pool, x, i, j):
    """Test y calculation against vyper implementation"""
    assume(i != j)

    balances = [vyper_metapool.balances(i) for i in range(2)]
    rates = [vyper_metapool.rates(i) for i in range(2)]
    virtual_balances = convert_to_virtual_balances(rates, balances)

    # need `eval` since this function is internal
    expected_y = vyper_metapool.eval(f"self.get_y({i}, {j}, {x}, {virtual_balances})")

    python_metapool = initialize_metapool(vyper_metapool, vyper_3pool)
    y = python_metapool.get_y(i, j, x, virtual_balances)
    assert y == expected_y


@given(
    st.integers(min_value=10 * D_UNIT, max_value=10**6 * D_UNIT),
    st.integers(min_value=0, max_value=1),
    st.integers(min_value=0, max_value=1),
)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_get_y_D(vyper_metapool, vyper_3pool, dx, i, j):
    """Test y calculation against vyper implementation"""
    assume(i != j)

    python_metapool = initialize_metapool(vyper_metapool, vyper_3pool)
    A = python_metapool.A
    D = python_metapool.D()
    rates = python_metapool.rates
    balances = python_metapool.balances
    vyper_balances = [vyper_metapool.balances(i) for i in range(2)]
    assert balances == vyper_balances
    virtual_balances = convert_to_virtual_balances(rates, balances)

    virtual_balances[j] += dx

    # Need `eval` since this function is internal.
    # `get_y_D` also takes in A_precise not A.
    A_precise = A * 100
    expected_y = vyper_metapool.eval(
        f"self.get_y_D({A_precise}, {i}, {virtual_balances}, {D})"
    )

    y = python_metapool.get_y_D(A, i, virtual_balances, D)
    assert y == expected_y


@given(positive_balance, positive_balance)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_calc_token_amount(vyper_metapool, vyper_3pool, x0, x1):
    """Test `calc_token_amount` against vyper implementation."""
    python_metapool = initialize_metapool(vyper_metapool, vyper_3pool)

    _balances = [x0, x1]
    rates = [vyper_metapool.rates(i) for i in range(len(_balances))]
    balances = convert_to_real_balances(rates, _balances)

    expected_lp_amount = vyper_metapool.calc_token_amount(balances, True)
    lp_amount = python_metapool.calc_token_amount(balances)

    assert lp_amount == expected_lp_amount


@given(positive_balance, positive_balance)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_add_liquidity(vyper_metapool, vyper_3pool, x0, x1):
    """Test `add_liquidity` against vyper implementation."""
    python_metapool = initialize_metapool(vyper_metapool, vyper_3pool)

    _balances = [x0, x1]
    rates = [vyper_metapool.rates(i) for i in range(len(_balances))]
    amounts = convert_to_real_balances(rates, _balances)

    old_vyper_balances = [vyper_metapool.balances(i) for i in range(len(_balances))]
    assert python_metapool.balances == old_vyper_balances

    expected_lp_amount = vyper_metapool.add_liquidity(amounts, 0)
    expected_balances = [vyper_metapool.balances(i) for i in range(len(_balances))]
    expected_lp_supply = vyper_metapool.totalSupply()

    lp_amount = python_metapool.add_liquidity(amounts)

    assert lp_amount == expected_lp_amount
    assert python_metapool.tokens == expected_lp_supply
    assert python_metapool.balances == expected_balances


@given(
    positive_balance,
    st.integers(min_value=0, max_value=1),
    st.integers(min_value=0, max_value=1),
)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_exchange(vyper_metapool, vyper_3pool, dx, i, j):
    """Test `exchange` against vyper implementation."""
    assume(i != j)

    python_metapool = initialize_metapool(vyper_metapool, vyper_3pool)

    old_vyper_balances = [vyper_metapool.balances(i) for i in range(2)]
    balances = python_metapool.balances
    assert balances == old_vyper_balances

    # convert to real units
    dx = dx * 10**18 // vyper_metapool.rates(i)

    expected_dy = vyper_metapool.exchange(i, j, dx, 0)
    dy, _ = python_metapool.exchange(i, j, dx)

    assert dy == expected_dy

    expected_balances = [vyper_metapool.balances(i) for i in range(2)]
    new_balances = python_metapool.balances
    assert new_balances == expected_balances


@given(
    positive_balance,
    st.integers(min_value=0, max_value=3),
    st.integers(min_value=0, max_value=3),
)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_exchange_underlying(vyper_metapool, vyper_3pool, dx, i, j):
    """Test `exchange_underlying` against vyper implementation."""
    assume(i != j)

    python_metapool = initialize_metapool(vyper_metapool, vyper_3pool)
    python_basepool = python_metapool.basepool

    # check metapool balances
    old_vyper_balances = [vyper_metapool.balances(i) for i in range(2)]
    balances = python_metapool.balances
    assert balances == old_vyper_balances
    # check basepool balances
    old_vyper_balances = [vyper_3pool.balances(i) for i in range(3)]
    balances = python_basepool.balances
    assert balances == old_vyper_balances

    # convert to real units
    if i == 0:
        dx = dx * 10**18 // vyper_metapool.rates(0)
    else:
        base_i = i - 1
        dx = dx * 10**18 // vyper_3pool.rates(base_i)

    expected_dy = vyper_metapool.exchange_underlying(i, j, dx, 0)
    dy, _ = python_metapool.exchange_underlying(i, j, dx)

    assert dy == expected_dy

    # check metapool balances
    expected_balances = [vyper_metapool.balances(i) for i in range(2)]
    new_balances = python_metapool.balances
    assert new_balances == expected_balances
    # check basepool balances
    expected_balances = [vyper_3pool.balances(i) for i in range(3)]
    new_balances = python_basepool.balances
    assert new_balances == expected_balances


@given(positive_balance, st.integers(min_value=0, max_value=1))
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_calc_withdraw_one_token(vyper_metapool, vyper_3pool, amount, i):
    """Test `calc_withdraw_one_coin` against vyper implementation."""
    assume(amount < vyper_metapool.totalSupply())

    python_metapool = initialize_metapool(vyper_metapool, vyper_3pool)

    expected_coin_amount = vyper_metapool.calc_withdraw_one_coin(amount, i)
    coin_amount, _ = python_metapool.calc_withdraw_one_coin(amount, i)
    assert coin_amount == expected_coin_amount


@given(positive_balance, st.integers(min_value=0, max_value=1))
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_remove_liquidity_one_coin(vyper_metapool, vyper_3pool, amount, i):
    """Test `remove_liquidity_one_coin` against vyper implementation."""
    assume(amount < vyper_metapool.totalSupply())

    python_metapool = initialize_metapool(vyper_metapool, vyper_3pool)

    old_vyper_balances = [vyper_metapool.balances(i) for i in range(2)]
    balances = python_metapool.balances
    assert balances == old_vyper_balances

    old_vyper_supply = vyper_metapool.totalSupply()
    lp_supply = python_metapool.tokens
    assert lp_supply == old_vyper_supply

    vyper_metapool.remove_liquidity_one_coin(amount, i, 0)
    expected_coin_balance = vyper_metapool.balances(i)
    expected_lp_supply = vyper_metapool.totalSupply()

    python_metapool.remove_liquidity_one_coin(amount, i)
    coin_balance = python_metapool.balances[i]
    lp_supply = python_metapool.tokens

    assert coin_balance == expected_coin_balance
    assert lp_supply == expected_lp_supply


@pytest.mark.skip(reason="WIP: various issues affecting reliabililty of test")
@given(
    positive_balance,
    positive_balance,
    positive_balance,
    positive_balance,
    positive_balance,
    st.integers(min_value=0, max_value=3),
    st.integers(min_value=0, max_value=3),
)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=10,
    deadline=None,
)
def test_dydxfee(x0, x1, b0, b1, b2, i, j):
    """Test `dydxfee` against discrete pricing using `exchange`"""
    # don't test between base underlyers here; will be done in regular pool tests
    assume(i != j and (i == 0 or j == 0))

    admin_fee = 5 * 10**9

    base_A = 3000
    base_rates = [10**18, 10**30, 10**30]
    base_balances = [b0, b1, b2]
    base_balances = convert_to_real_balances(base_rates, base_balances)
    base_n = len(base_balances)
    base_tokens = sum(base_balances)
    base_fee = 1 * 10**6
    base_fee = 0
    basepool = CurvePool(
        base_A,
        D=base_balances,
        n=base_n,
        rates=base_rates,
        tokens=base_tokens,
        fee=base_fee,
        admin_fee=admin_fee,
    )

    A = 1365
    rate_multiplier = 10**18
    p = [rate_multiplier, 10**18]
    balances = [x0, x1]
    balances = convert_to_real_balances(p, balances)
    n = len(balances)
    tokens = sum(balances)
    fee = 4 * 10**6
    fee = 0
    metapool = CurveMetaPool(
        A,
        D=balances,
        n=n,
        basepool=basepool,
        rate_multiplier=rate_multiplier,
        tokens=tokens,
        fee=fee,
        admin_fee=admin_fee,
    )

    _dydx = metapool.dydxfee(i, j)

    rates = [rate_multiplier] + base_rates
    _dx = 10**12
    dx = _dx * 10**18 // rates[i]
    dy, _ = metapool.exchange_underlying(i, j, dx)
    _dy = dy * rates[j] // 10**18

    price = _dy / _dx
    # print("\n")
    # print("Discretized derivative:", price)
    # print("Continuous derivative:", _dydx)
    # print(f"dx: {dx}, i: {i} ")
    # print(f"dy: {dy}, j: {j} ")
    # print(base_balances)
    # print(balances)
    # print("-------------------------------")

    assert _dydx > 0
    assert price > 0

    tol = 1e-12
    if i == 0:
        # due to our fee approximation for the continuous pricing case,
        # we won't get as much precision with using `exchange_underlying`,
        # which will use the exact fee logic in the contracts.
        tol = 0.0015

    assert abs(price - _dydx) < tol



__all__ = ["PoolMetaData", "PoolMetaDataInterface"]


from curvesim.exceptions import CurvesimException
from curvesim.network.subgraph import has_redemption_prices
from curvesim.pool.cryptoswap import CurveCryptoPool
from curvesim.pool.sim_interface import (
    SimCurveCryptoPool,
    SimCurveMetaPool,
    SimCurvePool,
    SimCurveRaiPool,
)
from curvesim.pool.stableswap import CurveMetaPool, CurvePool, CurveRaiPool
from curvesim.pool_data.metadata.cryptoswap import CryptoswapMetaData

from .base import PoolMetaDataInterface
from .stableswap import StableswapMetaData

_SIM_POOL_TYPE = {
    CurvePool: SimCurvePool,
    CurveMetaPool: SimCurveMetaPool,
    CurveRaiPool: SimCurveRaiPool,
    CurveCryptoPool: SimCurveCryptoPool,
}

_METADATA_TYPE = {
    CurvePool: StableswapMetaData,
    CurveMetaPool: StableswapMetaData,
    CurveRaiPool: StableswapMetaData,
    CurveCryptoPool: CryptoswapMetaData,
}


def PoolMetaData(metadata_dict):
    """
    A factory function to return a :class:`PoolMetaDataInterface`.

    Parameters
    ----------
    metadata_dict : dict
        Pool metadata in the format returned by
        :func:`curvesim.network.subgraph.pool_snapshot`.

    Returns
    -------
    :class:`PoolMetaDataInterface`
    """
    metadata_type, pool_type, sim_pool_type = _parse_metadata_for_types(metadata_dict)
    return metadata_type(metadata_dict, pool_type, sim_pool_type)


def _parse_metadata_for_types(metadata_dict):
    if metadata_dict["basepool"]:
        if _has_redemption_prices(metadata_dict):
            pool_type = CurveRaiPool
        else:
            pool_type = CurveMetaPool
    else:
        if metadata_dict["params"].get("gamma"):
            pool_type = CurveCryptoPool
        else:
            pool_type = CurvePool

    try:
        sim_pool_type = _SIM_POOL_TYPE[pool_type]
    except KeyError as e:
        raise CurvesimException(
            f"No sim pool type for this pool type: {pool_type}"
        ) from e
    try:
        metadata_type = _METADATA_TYPE[pool_type]
    except KeyError as e:
        raise CurvesimException(
            f"No metadata type for this pool type: {pool_type}"
        ) from e

    return metadata_type, pool_type, sim_pool_type


def _has_redemption_prices(metadata_dict):
    address = metadata_dict["address"]
    chain = metadata_dict["chain"]
    return has_redemption_prices(address, chain)



from curvesim.logging import get_logger
from curvesim.metrics.state_log import StateLog
from curvesim.templates import Strategy

from .trader import SimpleArbitrageur

logger = get_logger(__name__)


class SimpleStrategy(Strategy):  # pylint: disable=too-few-public-methods
    """
    Class Attributes
    ----------------
    trader_class : :class:`~curvesim.pipelines.simple.trader.SimpleArbitrageur`
        Class for creating trader instances.
    state_log_class : :class:`~curvesim.metrics.StateLog`
        Class for creating state logger instances.
    """

    trader_class = SimpleArbitrageur
    state_log_class = StateLog

    def _get_trader_inputs(self, sample):
        return (sample.prices,)



from abc import ABC, abstractmethod


class ResultPlotter(ABC):
    """
    Result plotter base class with required properties for any result plotter object.

    """

    def __init__(self, plot_data, plot_results, plot_summary):
        """
        Parameters
        ----------
        plot_data : callable
            A function that takes a :class:`.metrics.SimResults` object and plots the
            data returned by :func:`SimResults.data`.

        plot_results : callable
            A function that takes a :class:`.metrics.SimResults` object and plots the
            data returned by both :func:`SimResults.summary` and :func:`SimResults.data`

        plot_summary : callable
            A function that takes a :class:`.metrics.SimResults` object and plots the
            data returned by :func:`SimResults.summary`.
        """

        self.plot_data = plot_data
        self.plot_results = plot_results
        self.plot_summary = plot_summary

    def plot(self, results, summary=True, data=True, save_as=None):
        """
        Returns and optionally saves a plot of the results data.
        Used in :func:`.results.SimResults.plot`

        Parameters
        ----------
        summary : bool, default=True
            If true, includes summary data in the plot.

        data : bool, default=True
            If true, includes timeseries data in the plot.

        save_as : str, optional
            Path to save plot output to.


        Returns
        -------
        A chart object.

        """
        if summary and data:
            chart = self.plot_results(results)
        elif summary:
            chart = self.plot_summary(results)
        elif data:
            chart = self.plot_data(results)

        if save_as:
            self.save(chart, save_as)

        return chart

    @abstractmethod
    def save(self, chart, save_as):
        """
        Saves the chart output by the plot() method.

        Parameters
        ----------
        chart
            A chart object.

        save_as : str
            Path to save plot output to.
        """
        raise NotImplementedError



from curvesim.pool_data import get_metadata

from curvesim.pool.sim_interface import SimCurveMetaPool, SimCurvePool
from curvesim.pool.stableswap import CurveMetaPool, CurvePool


properties = ["chain", "pool_type", "sim_pool_type", "coins", "coin_names", "n"]

# To generate stored properties for a pool:
# stored_properties = {}
# for p in properties:
#      stored_properties[p] = getattr(metadata, p)


def _test_pool_metadata(address, chain, stored_properties, stored_dict):
    metadata = get_metadata(address, chain)
    metadata_dict = metadata._dict

    # Remove non-static items from dict
    del metadata_dict["timestamp"], metadata_dict["reserves"], metadata_dict["params"]

    if metadata_dict["basepool"] is not None:
        del (
            metadata_dict["basepool"]["timestamp"],
            metadata_dict["basepool"]["reserves"],
            metadata_dict["basepool"]["params"],
        )

    for p in properties:
        assert stored_properties[p] == getattr(
            metadata, p
        ), f"Metadata property '{p}' did not match for {chain}:{address}"

    assert (
        metadata_dict == stored_dict
    ), f"Metadata dict did not match for {chain}:{address}"


def test_lending_pool_metadata():
    # a3crv
    address = "0xDeBF20617708857ebe4F679508E7b7863a8A8EeE"
    chain = "mainnet"

    stored_properties = {
        "chain": "mainnet",
        "pool_type": CurvePool,
        "sim_pool_type": SimCurvePool,
        "coins": [
            "0x6B175474E89094C44Da98b954EedeAC495271d0F",
            "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48",
            "0xdAC17F958D2ee523a2206206994597C13D831ec7",
        ],
        "coin_names": ["DAI", "USDC", "USDT"],
        "n": 3,
    }

    stored_dict = {
        "name": "Curve.fi aDAI/aUSDC/aUSDT",
        "address": "0xDeBF20617708857ebe4F679508E7b7863a8A8EeE",
        "chain": "mainnet",
        "symbol": "a3CRV",
        "version": 1,
        "pool_type": "LENDING",
        "basepool": None,
        "coins": {
            "names": ["DAI", "USDC", "USDT"],
            "addresses": [
                "0x6B175474E89094C44Da98b954EedeAC495271d0F",
                "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48",
                "0xdAC17F958D2ee523a2206206994597C13D831ec7",
            ],
            "decimals": [18, 6, 6],
            "wrapper": {
                "names": ["aDAI", "aUSDC", "aUSDT"],
                "addresses": [
                    "0x028171bCA77440897B824Ca71D1c56caC55b68A3",
                    "0xBcca60bB61934080951369a648Fb03DF4F96263C",
                    "0x3Ed3B47Dd13EC9a98b44e6204A523E766B225811",
                ],
                "decimals": [18, 6, 6],
            },
        },
    }

    _test_pool_metadata(address, chain, stored_properties, stored_dict)


def test_l2_pool_metadata():
    # 2crv
    address = "0x7f90122BF0700F9E7e1F688fe926940E8839F353"
    chain = "arbitrum"

    stored_properties = {
        "chain": "arbitrum",
        "pool_type": CurvePool,
        "sim_pool_type": SimCurvePool,
        "coins": [
            "0xFF970A61A04b1cA14834A43f5dE4533eBDDB5CC8",
            "0xFd086bC7CD5C481DCC9C85ebE478A1C0b69FCbb9",
        ],
        "coin_names": ["USDC", "USDT"],
        "n": 2,
    }

    stored_dict = {
        "name": "Curve.fi USDC/USDT",
        "address": "0x7f90122BF0700F9E7e1F688fe926940E8839F353",
        "chain": "arbitrum",
        "symbol": "2CRV",
        "version": 1,
        "pool_type": "REGISTRY_V1",
        "basepool": None,
        "coins": {
            "names": ["USDC", "USDT"],
            "addresses": [
                "0xFF970A61A04b1cA14834A43f5dE4533eBDDB5CC8",
                "0xFd086bC7CD5C481DCC9C85ebE478A1C0b69FCbb9",
            ],
            "decimals": [6, 6],
        },
    }

    _test_pool_metadata(address, chain, stored_properties, stored_dict)


def test_mainnet_metapool_metadata():
    # OUSD
    address = "0x87650d7bbfc3a9f10587d7778206671719d9910d"
    chain = "mainnet"

    stored_properties = {
        "chain": "mainnet",
        "pool_type": CurveMetaPool,
        "sim_pool_type": SimCurveMetaPool,
        "coins": [
            "0x2A8e1E676Ec238d8A992307B495b45B3fEAa5e86",
            "0x6B175474E89094C44Da98b954EedeAC495271d0F",
            "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48",
            "0xdAC17F958D2ee523a2206206994597C13D831ec7",
        ],
        "coin_names": ["OUSD", "DAI", "USDC", "USDT"],
        "n": [2, 3],
    }

    stored_dict = {
        "name": "Curve.fi Factory USD Metapool: Origin Dollar",
        "address": "0x87650D7bbfC3A9F10587d7778206671719d9910D",
        "chain": "mainnet",
        "symbol": "OUSD3CRV-f",
        "version": 1,
        "pool_type": "METAPOOL_FACTORY",
        "basepool": {
            "name": "Curve.fi DAI/USDC/USDT",
            "address": "0xbEbc44782C7dB0a1A60Cb6fe97d0b483032FF1C7",
            "chain": "mainnet",
            "symbol": "3Crv",
            "version": 1,
            "pool_type": "REGISTRY_V1",
            "coins": {
                "names": ["DAI", "USDC", "USDT"],
                "addresses": [
                    "0x6B175474E89094C44Da98b954EedeAC495271d0F",
                    "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48",
                    "0xdAC17F958D2ee523a2206206994597C13D831ec7",
                ],
                "decimals": [18, 6, 6],
            },
            "basepool": None,
        },
        "coins": {
            "names": ["OUSD", "3Crv"],
            "addresses": [
                "0x2A8e1E676Ec238d8A992307B495b45B3fEAa5e86",
                "0x6c3F90f043a72FA612cbac8115EE7e52BDe6E490",
            ],
            "decimals": [18, 18],
        },
    }

    _test_pool_metadata(address, chain, stored_properties, stored_dict)



"""
These mixins extend Curve pool classes for the purpose of generating iterables
of pool instances configured by different simulation parameters.

Each mixin defines a different pool type and a set of special attribute setters.
"""

from curvesim.pool.cryptoswap.calcs import newton_D
from curvesim.pool.sim_interface import (
    SimCurveCryptoPool,
    SimCurveMetaPool,
    SimCurvePool,
)


class CurvePoolMixin:
    """
    Parameter sampler mixin for Curve stableswap pools.
    Defines special attribute setters.
    """

    @property
    def _pool_type(self):
        return SimCurvePool

    @property
    def setters(self):
        """
        Returns
        -------
        dict
            A dictionary containing the special setters for the pool parameters.
        """
        return {"D": stableswap_D_to_balances}


class CurveMetaPoolMixin:
    """
    Parameter sampler mixin for Curve stableswap meta-pools.
    Defines special attribute setters.
    """

    @property
    def _pool_type(self):
        return SimCurveMetaPool

    @property
    def setters(self):
        """
        Returns
        -------
        dict
            A dictionary containing the special setters for the pool parameters.
        """
        return {"D": stableswap_D_to_balances, "D_base": stableswap_D_base_to_balances}


class CurveCryptoPoolMixin:
    """
    Parameter sampler mixin for Curve cryptoswap pools.
    Defines special attribute setters.
    """

    @property
    def _pool_type(self):
        return SimCurveCryptoPool

    @property
    def setters(self):
        """
        Returns
        -------
        dict
            A dictionary containing the special setters for the pool parameters.
        """
        return {
            "D": cryptoswap_D_to_balances,
            "A": set_cryptoswap_A,
            "gamma": set_cryptoswap_gamma,
        }


def set_cryptoswap_A(pool, A):
    """
    Sets A on the pool, with the following side-effects:
    - recalculate and cache D
    - recalculate and cache virtual_price

    Parameters
    ----------
    pool : instance of _pool_type
    A : int
        The A parameter
    """
    xp = pool._xp()  # pylint: disable=protected-access
    gamma = pool.gamma
    D = newton_D(A, gamma, xp)
    pool.D = D
    pool.A = A
    virtual_price = pool.get_virtual_price()
    pool.virtual_price = virtual_price


def set_cryptoswap_gamma(pool, gamma):
    """
    Sets gamma on the pool, with the following side-effects:
    - recalculate and cache D
    - recalculate and cache virtual_price

    Parameters
    ----------
    pool : instance of _pool_type
    gamma : int
        The gamma parameter
    """
    xp = pool._xp()  # pylint: disable=protected-access
    A = pool.A
    D = newton_D(A, gamma, xp)
    pool.D = D
    pool.gamma = gamma
    virtual_price = pool.get_virtual_price()
    pool.virtual_price = virtual_price


def stableswap_D_to_balances(pool, D):
    """
    Sets the balance for each token in the pool based on the provided
    invariant value.

    Parameters
    ----------
    pool : instance of _pool_type
    D : int
        The invariant value.
    """
    rates = pool.rates
    n = pool.n
    pool.balances = [D // n * 10**18 // r for r in rates]


def stableswap_D_base_to_balances(pool, D_base):
    """
    Sets the balance for each token in the basepool based on the provided
    invariant value.

    Parameters
    ----------
    pool : instance of _pool_type
    D_base : int
        The invariant value for the basepool.
    """
    basepool = pool.basepool
    rates = basepool.rates
    n = basepool.n
    basepool.balances = [D_base // n * 10**18 // r for r in rates]


def cryptoswap_D_to_balances(pool, D):
    """
    Sets the balance for each token in the pool based on the provided
    invariant value.

    Parameters
    ----------
    pool : instance of _pool_type
    D : int
        The invariant value.
    """
    pool.D = D
    pool.balances = pool._convert_D_to_balances(D)  # pylint: disable=protected-access



"""Unit tests for CurvePool"""
from hypothesis import HealthCheck, assume, given, settings
from hypothesis import strategies as st

from curvesim.pool import CurvePool


def initialize_pool(vyper_pool):
    """
    Initialize python-based pool from the state variables of the
    vyper-based implementation.
    """
    A = vyper_pool.A()
    n_coins = vyper_pool.N_COINS()
    balances = [vyper_pool.balances(i) for i in range(n_coins)]
    rates = [vyper_pool.rates(i) for i in range(n_coins)]
    lp_total_supply = vyper_pool.totalSupply()
    fee = vyper_pool.fee()
    admin_fee = vyper_pool.admin_fee()
    pool = CurvePool(
        A,
        D=balances,
        n=n_coins,
        rates=rates,
        tokens=lp_total_supply,
        fee=fee,
        admin_fee=admin_fee,
    )
    return pool


def test_get_D_against_prod(vyper_3pool, mainnet_3pool_state):
    """
    Test boa value against live contract.

    This checks boa is working correctly and also ensures mainnet
    state stays consistent.
    """
    # Compare against virtual price since that's exposed externally
    # while `get_D` is internal in the contract.
    D = vyper_3pool.D()
    total_supply = mainnet_3pool_state["lp_tokens"]
    virtual_price = D * 10**18 // total_supply

    expected_virtual_price = mainnet_3pool_state["virtual_price"]
    assert virtual_price == expected_virtual_price


def test_get_D_mainnet(vyper_3pool):
    """
    Test D calculation against vyper implementation using
    mainnet state.
    """
    expected_D = vyper_3pool.D()

    python_3pool = initialize_pool(vyper_3pool)
    D = python_3pool.D()

    assert D == expected_D


# We can assume the contract works on more extreme values; we only need
# to be reasonably certain our results are consistent, so we can check
# a smaller range.
#
# With 18 decimal precision, it seems reasonable to pick these bounds
D_UNIT = 10**18
positive_balance = st.integers(min_value=10**5 * D_UNIT, max_value=10**10 * D_UNIT)


@given(positive_balance, positive_balance, positive_balance)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_get_D(vyper_3pool, x0, x1, x2):
    """Test D calculation against vyper implementation."""

    _balances = [x0, x1, x2]
    p = [vyper_3pool.rates(i) for i in range(len(_balances))]
    balances = [x * 10**18 // p for x, p in zip(_balances, p)]

    vyper_3pool.eval(f"self.balances={balances}")
    expected_D = vyper_3pool.D()

    pool = initialize_pool(vyper_3pool)
    D = pool.D()

    assert D == expected_D


def test_get_D_balanced():
    """Sanity check for when pool is perfectly balanced"""

    # create balanced pool
    balances = [
        295949605740077000000000000,
        295949605740077,
        295949605740077,
    ]
    rates = [10**18, 10**30, 10**30]
    n_coins = 3
    A = 5858

    pool = CurvePool(A, D=balances, n=n_coins, rates=rates)
    D = pool.D()

    virtualized_balances = [b * p // 10**18 for b, p in zip(balances, rates)]
    expected_D = sum(virtualized_balances)

    assert D == expected_D


def test_get_virtual_price(vyper_3pool):
    """Test `get_virtual_price` against vyper implementation."""

    python_3pool = initialize_pool(vyper_3pool)
    virtual_price = python_3pool.get_virtual_price()
    expected_virtual_price = vyper_3pool.get_virtual_price()
    assert virtual_price == expected_virtual_price


def test_get_y(vyper_3pool, mainnet_3pool_state):
    """Test y calculation against vyper implementation"""

    virtual_balances = mainnet_3pool_state["virtual_balances"]

    i = 0
    j = 1
    x = 516 * 10**18
    # need `eval` since this function is internal
    expected_y = vyper_3pool.eval(f"self.get_y({i}, {j}, {x}, {virtual_balances})")

    python_3pool = initialize_pool(vyper_3pool)
    y = python_3pool.get_y(i, j, x, virtual_balances)
    assert y == expected_y


def test_get_y_D(vyper_3pool):
    """Test y calculation against vyper implementation"""

    python_3pool = initialize_pool(vyper_3pool)
    A = python_3pool.A
    virtual_balances = python_3pool._xp()  # pylint: disable=protected-access
    D = python_3pool.D()

    i = 0
    j = 1
    dx = 516 * 10**18
    virtual_balances[j] += dx
    expected_y = vyper_3pool.eval(f"self.get_y_D({A}, {i}, {virtual_balances}, {D})")

    y = python_3pool.get_y_D(A, i, virtual_balances, D)
    assert y == expected_y


@given(positive_balance, positive_balance, positive_balance)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_calc_token_amount(vyper_3pool, x0, x1, x2):
    """Test `calc_token_amount` against vyper implementation."""
    python_3pool = initialize_pool(vyper_3pool)

    _balances = [x0, x1, x2]
    rates = [vyper_3pool.rates(i) for i in range(len(_balances))]
    balances = [b * 10**18 // r for b, r in zip(_balances, rates)]

    expected_lp_amount = vyper_3pool.calc_token_amount(balances, True)
    lp_amount = python_3pool.calc_token_amount(balances)

    assert lp_amount == expected_lp_amount


@given(positive_balance, positive_balance, positive_balance)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_add_liquidity(vyper_3pool, x0, x1, x2):
    """Test `add_liquidity` against vyper implementation."""
    python_3pool = initialize_pool(vyper_3pool)

    _balances = [x0, x1, x2]
    rates = [vyper_3pool.rates(i) for i in range(len(_balances))]
    amounts = [b * 10**18 // r for b, r in zip(_balances, rates)]

    old_vyper_balances = [vyper_3pool.balances(i) for i in range(len(_balances))]
    assert python_3pool.balances == old_vyper_balances

    expected_lp_amount = vyper_3pool.add_liquidity(amounts, 0)
    expected_balances = [vyper_3pool.balances(i) for i in range(len(_balances))]
    expected_lp_supply = vyper_3pool.totalSupply()

    lp_amount = python_3pool.add_liquidity(amounts)

    assert lp_amount == expected_lp_amount
    assert python_3pool.tokens == expected_lp_supply
    assert python_3pool.balances == expected_balances


@given(
    positive_balance,
    st.integers(min_value=0, max_value=2),
    st.integers(min_value=0, max_value=2),
)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_exchange(vyper_3pool, dx, i, j):
    """Test `exchange` against vyper implementation."""
    assume(i != j)

    python_3pool = initialize_pool(vyper_3pool)

    old_vyper_balances = [vyper_3pool.balances(i) for i in range(3)]
    balances = python_3pool.balances
    assert balances == old_vyper_balances

    # convert to real units
    dx = dx * 10**18 // vyper_3pool.rates(i)

    expected_dy = vyper_3pool.exchange(i, j, dx, 0)
    dy, _ = python_3pool.exchange(i, j, dx)

    assert dy == expected_dy

    expected_balances = [vyper_3pool.balances(i) for i in range(3)]
    new_balances = python_3pool.balances
    assert new_balances == expected_balances


@given(positive_balance, st.integers(min_value=0, max_value=2))
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_calc_withdraw_one_token(vyper_3pool, amount, i):
    """Test `calc_withdraw_one_coin` against vyper implementation."""
    assume(amount < vyper_3pool.totalSupply())

    python_3pool = initialize_pool(vyper_3pool)

    expected_coin_amount = vyper_3pool.calc_withdraw_one_coin(amount, i)
    coin_amount, _ = python_3pool.calc_withdraw_one_coin(amount, i)
    assert coin_amount == expected_coin_amount


@given(positive_balance, st.integers(min_value=0, max_value=2))
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_remove_liquidity_one_coin(vyper_3pool, amount, i):
    """Test `remove_liquidity_one_coin` against vyper implementation."""
    assume(amount < vyper_3pool.totalSupply())

    python_3pool = initialize_pool(vyper_3pool)

    old_vyper_balances = [vyper_3pool.balances(i) for i in range(3)]
    balances = python_3pool.balances
    assert balances == old_vyper_balances

    old_vyper_supply = vyper_3pool.totalSupply()
    lp_supply = python_3pool.tokens
    assert lp_supply == old_vyper_supply

    vyper_3pool.remove_liquidity_one_coin(amount, i, 0)
    expected_coin_balance = vyper_3pool.balances(i)
    expected_lp_supply = vyper_3pool.totalSupply()

    python_3pool.remove_liquidity_one_coin(amount, i)
    coin_balance = python_3pool.balances[i]
    lp_supply = python_3pool.tokens

    assert coin_balance == expected_coin_balance
    assert lp_supply == expected_lp_supply



"""Module to house the `SimPool` extension of the `CurveCryptoPool`."""
from math import prod

from curvesim.exceptions import SimPoolError
from curvesim.templates import SimAssets
from curvesim.templates.sim_pool import SimPool
from curvesim.utils import cache, override

from ..cryptoswap import CurveCryptoPool
from ..cryptoswap.calcs import newton_D
from ..cryptoswap.calcs.factory_2_coin import _sqrt_int
from ..cryptoswap.calcs.tricrypto_ng import _cbrt
from .asset_indices import AssetIndicesMixin


class SimCurveCryptoPool(SimPool, AssetIndicesMixin, CurveCryptoPool):
    """
    Class to enable use of CurveCryptoPool in simulations by exposing
    a generic interface (`SimPool`).
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        precisions = self.precisions  # pylint: disable=no-member
        for p in precisions:
            if p != 1:
                raise SimPoolError(
                    "SimPool must have 18 decimals (precision 1) for each coin."
                )

    @property
    @override
    @cache
    def asset_names(self):
        """Return list of asset names."""
        return self.coin_names

    @property
    @override
    def _asset_balances(self):
        """Return list of asset balances in same order as asset_names."""
        return self.balances

    @override
    def price(self, coin_in, coin_out, use_fee=True):
        """
        Returns the spot price of `coin_in` quoted in terms of `coin_out`,
        i.e. the ratio of output coin amount to input coin amount for
        an "infinitesimally" small trade.

        Coin IDs should be strings but as a legacy feature integer indices
        corresponding to the pool implementation are allowed (caveat lector).

        Parameters
        ----------
        coin_in : str, int
            ID of coin to be priced; in a swapping context, this is
            the "in"-token.
        coin_out : str, int
            ID of quote currency; in a swapping context, this is the
            "out"-token.
        use_fee: bool, default=True
            Deduct fees.

        Returns
        -------
        float
            Price of `coin_in` quoted in `coin_out`
        """
        i, j = self.get_asset_indices(coin_in, coin_out)
        p = self.dydx(i, j, use_fee=use_fee)
        return p

    @override
    def trade(self, coin_in, coin_out, size):
        """
        Perform an exchange between two coins.

        Coin IDs should be strings but as a legacy feature integer indices
        corresponding to the pool implementation are allowed (caveat lector).

        Note that all amounts are normalized to be in the same units as
        pool value, i.e. `XCP`.  This simplifies cross-token comparisons
        and creation of metrics.


        Parameters
        ----------
        coin_in : str, int
            ID of "in" coin.
        coin_out : str, int
            ID of "out" coin.
        size : int
            Amount of coin `i` being exchanged.

        Returns
        -------
        (int, int)
            (amount of coin `j` received, trading fee)
        """
        i, j = self.get_asset_indices(coin_in, coin_out)
        amount_out, fee = self.exchange(i, j, size)
        return amount_out, fee

    @override
    def get_max_trade_size(self, coin_in, coin_out, out_balance_perc=0.15):
        """
        Get the approximate in-amount to achieve the given percentage
        of the out-token balance.

        Parameters
        ----------
        coin_in: int
            name or index of in-token
        coin_out: int
            name or index of out-token
        out_balance_perc : float
            percentage of the out-token balance that should remain after swap

        Returns
        -------
        int
            An approximate quantity to swap to achieve the target out-token
            balance
        """
        i, j = self.get_asset_indices(coin_in, coin_out)

        xp = self._xp()
        xp_j = int(xp[j] * out_balance_perc)

        in_amount = self.get_y(j, i, xp_j, xp) - xp[i]
        if i > 0:
            in_amount = in_amount * 10**18 // self.price_scale[i - 1]
        return in_amount

    @override
    def get_min_trade_size(self, coin_in):
        """
        Return the minimal trade size allowed for the pool.

        Parameters
        ----------
        coin_in : str, int
            ID of "in" coin.

        Returns
        -------
        int
            The minimal trade size
        """
        (i,) = self.get_asset_indices(coin_in)
        min_amount = 10**18
        if i > 0:
            min_amount = min_amount * 10**18 // self.price_scale[i - 1]
        return min_amount

    @override
    def prepare_for_trades(self, timestamp):
        """
        Updates the pool's _block_timestamp attribute to current sim time.

        Parameters
        ----------
        timestamp : datetime.datetime
            The current timestamp in the simulation.
        """

        timestamp = int(timestamp.timestamp())  # unix timestamp in seconds
        self._increment_timestamp(timestamp=timestamp)

    @override
    def prepare_for_run(self, prices):
        """
        Sets price parameters to the first simulation price and updates
        balances to be equally-valued.

        Balances are updated so that xcp(D) is preserved, but D may change.

        Parameters
        ----------
        timestamp : pandas.DataFrame
            The price time_series, price_sampler.prices.
        """
        xcp = self._get_xcp(self.D)
        n = self.n

        if n == 2:
            root = _sqrt_int
        elif n == 3:
            root = _cbrt

        # Get/set initial prices
        initial_prices = prices.iloc[0, 0 : n - 1].tolist()
        initial_prices = [int(10**18 / p) for p in initial_prices]

        self.last_prices = initial_prices
        self.price_scale = initial_prices
        self._price_oracle = initial_prices

        # Upbdate balances, preserving xcp
        initial_prices_root = [root(p) for p in initial_prices]
        new_D = prod(initial_prices_root) * xcp * n // 10 ** (18 * (n - 1))
        balances = self._convert_D_to_balances(new_D)
        self.balances = balances

        # Recompute and store D & virtual_price
        xp = self._xp()  # pylint: disable=protected-access
        computed_D = newton_D(self.A, self.gamma, xp)
        self.D = computed_D

        virtual_price = self.get_virtual_price()
        self.virtual_price = virtual_price

    @property
    @override
    @cache
    def assets(self):
        """
        Return :class:`.SimAssets` object with the properties of the pool's assets.

        Returns
        -------
        SimAssets
            SimAssets object that stores the properties of the pool's assets.
        """
        return SimAssets(self.coin_names, self.coin_addresses, self.chain)



"""
Tools for implementing and running simulation pipelines.

The basic model for a pipeline is demonstrated in the implementation of
:func:`run_pipeline`.  It takes in a `param_sampler`, `price_sampler`, and
`strategy`.

Pipelines iterate over pools with parameters set by
:mod:`curvesim.iterators.param_samplers` and time-series data produced by
:mod:`curvesim.iterators.price_samplers`.  At each timestemp, the
the :class:`~curvesim.pipelines.templates.Strategy` dictates what is done.

A typical pipeline implementation is a function taking in whatever market data needed;
pool data such as :class:`~curvesim.pool_data.metadata.PoolMetaDataInterface`;
instantiates a param_sampler, price_sampler, and strategy; and invokes `run_pipeline`,
returning its result metrics.
"""
from multiprocessing import Pool as cpu_pool

from curvesim.logging import (
    configure_multiprocess_logging,
    get_logger,
    multiprocessing_logging_queue,
)

logger = get_logger(__name__)


def run_pipeline(param_sampler, price_sampler, strategy, ncpu=4):
    """
    Core function for running pipelines.

    Typically called within a function specifying the pipeline components
    (see, e.g., :func:`curvesim.pipelines.vol_limited_arb.pipeline`)

    Parameters
    ----------
    param_sampler : iterator
        An iterator that returns pool parameters (see :mod:`.param_samplers`).

    price_sampler : iterator
        An iterator that returns (minimally) a time-series of prices
        (see :mod:`.price_samplers`).

    strategy: callable
        A function dictating what happens at each timestep.

    ncpu : int, default=4
        Number of cores to use.

    Returns
    -------
    results : tuple
        Contains the metrics produced by the strategy.

    """
    if ncpu > 1:
        with multiprocessing_logging_queue() as logging_queue:
            strategy_args_list = [
                (pool, params, price_sampler) for pool, params in param_sampler
            ]

            wrapped_args_list = [
                (strategy, logging_queue, *args) for args in strategy_args_list
            ]

            with cpu_pool(ncpu) as clust:
                results = tuple(
                    zip(*clust.starmap(wrapped_strategy, wrapped_args_list))
                )
                clust.close()
                clust.join()  # coverage needs this

    else:
        results = []
        for pool, params in param_sampler:
            metrics = strategy(pool, params, price_sampler)
            results.append(metrics)
        results = tuple(zip(*results))

    return results


def wrapped_strategy(strategy, logging_queue, *args):
    """
    This wrapper ensures we configure logging to use the
    multiprocessing enqueueing logic within the new process.

    Must be defined at the top-level of the module so it can
    be pickled.
    """
    configure_multiprocess_logging(logging_queue)
    return strategy(*args)



from curvesim.utils import get_pairs


# pylint: disable-next=too-few-public-methods
class SimAssets:
    """
    Stores the properties of the assets to be used in a simulation. Currently, only
    specific coins identified by their address/chain are supported. This will be
    expanded to "abstract" assets (e.g., some function of multiple coins) in the near
    future.
    """

    def __init__(self, symbols, addresses, chain):
        self.symbols = symbols
        self.addresses = addresses
        self.chain = chain
        self.symbol_pairs = get_pairs(symbols)
        self.address_pairs = get_pairs(addresses)



"""
Parameter overrides for cases that need special handling.
"""
from copy import deepcopy

subgraph_overrides = {
    "pool_snapshot": {
        ("0xA5407eAE9Ba41422680e2e00537571bcC53efBfD", "mainnet"): {
            "pool_type": "REGISTRY_V1"
        }
    }
}


def override_subgraph_data(data, function_key, data_key):
    """
    Applies overrides to subgraph data.

    Parameters
    ----------
    data : dict
        The data to be potentially overriden.

    function_key : str
        The function to lookup overrides for (i.e., subgraph_overrides[function_key]).

    data_key :
        The key for the specific item to lookup overrides for
        (i.e., subgraph_overrides[function_key][data_key])

    Returns
    -------
    dict
        The input data with any overrides applied.

    """
    overrides = subgraph_overrides[function_key]

    if data_key in overrides:
        data = deepcopy(data)
        for key, val in overrides[data_key].items():
            data.update({key: val})

    return data



"""
Simple health check for the package.

Also provides version info from the command-line.
"""
import argparse
import platform
import time

from .sim import autosim
from .version import __version__


def hello_world():
    """Simple sim run as a health check."""
    # pylint: disable=redefined-outer-name
    t = time.time()
    res = autosim("0xbEbc44782C7dB0a1A60Cb6fe97d0b483032FF1C7", test=True, ncpu=1)
    elapsed = time.time() - t
    print("Elapsed time:", elapsed)

    return res


def _python_info():
    """
    Return formatted string for python implementation and version.

    Returns
    --------
    str:
        Implementation name, version, and platform
    """
    impl = platform.python_implementation()
    version = platform.python_version()
    system = platform.system()
    return f"{impl} {version} on {system}"


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="curvesim",
        description="Simulate Curve pools in Python",
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {__version__}, {_python_info()}",
    )
    args = parser.parse_args()

    # `--version` option automatically exits, so we can
    # just run this.  If other options are added, we'll
    # need to check args and decide to run or not.
    res = hello_world()



__all__ = ["SimCurvePool", "SimCurveRaiPool", "SimCurveMetaPool", "SimCurveCryptoPool"]

from .cryptoswap import SimCurveCryptoPool
from .metapool import SimCurveMetaPool
from .pool import SimCurvePool
from .raipool import SimCurveRaiPool



"""
Various network connectors and related utilities.

Primary data sources for general exchange prices and volumes are Coingecko and Nomics.
Curve subgraph is used for fetching pool data.  The Llama Airforce REST API will also
be used in the future.

There is not a well-defined public API for this subpackage yet.  A lot of the code
has been converted to use `asyncio` from a legacy implementation and further,
substantive improvements are anticipated.
"""



from altair import Tooltip


def make_tooltip(encoding, metric_axis, factors, prefix=None):
    """Makes a tooltip for a subplot."""
    tooltip = []

    x_shorthand = encoding["x"]["shorthand"]
    if isinstance(x_shorthand, str) and "timestamp" in x_shorthand:
        tooltip.append(Tooltip(x_shorthand, title="Time"))

    title = encoding[metric_axis]["title"]
    if prefix:
        title = f"{prefix} {title}"

    tooltip.append(Tooltip(encoding[metric_axis]["shorthand"], title=title))

    return tooltip + factors



from curvesim.exceptions import CurvesimValueError, SimPoolError
from curvesim.templates import SimAssets
from curvesim.templates.sim_pool import SimPool
from curvesim.utils import cache, override

from ..stableswap import CurveMetaPool
from .asset_indices import AssetIndicesMixin


class SimCurveMetaPool(SimPool, AssetIndicesMixin, CurveMetaPool):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # The rates check has a couple special cases:
        # 1. For metapools, we need to use the basepool rates
        #    instead of the virtual price for the basepool.
        # 2. If `rate_multiplier` is passed as a kwarg, this will
        #    likely be some price, which we should skip.
        rates = [self.rate_multiplier] + self.basepool.rates
        if "rate_multiplier" in kwargs:
            rates = rates[1:]

        for r in rates:
            if r != 10**18:
                raise SimPoolError("SimPool must have 18 decimals for each coin.")

    @property
    @override
    @cache
    def asset_names(self):
        """
        Return list of asset names.

        For metapools, our convention is to place the basepool LP token last.
        """
        meta_coin_names = self.coin_names[:-1]
        base_coin_names = self.basepool.coin_names
        bp_token_name = self.coin_names[-1]

        return [*meta_coin_names, *base_coin_names, bp_token_name]

    @property
    @override
    def _asset_balances(self):
        """Return list of asset balances in same order as asset_names."""
        meta_balances = self.balances[:-1]
        base_balances = self.basepool.balances
        bp_token_balances = self.balances[-1]

        return [*meta_balances, *base_balances, bp_token_balances]

    @override
    def price(self, coin_in, coin_out, use_fee=True):
        """
        Returns the spot price of `coin_in` quoted in terms of `coin_out`,
        i.e. the ratio of output coin amount to input coin amount for
        an "infinitesimally" small trade.

        Coin IDs should be strings but as a legacy feature integer indices
        corresponding to the pool implementation are allowed (caveat lector).

        The indices are assumed to include base pool underlyer indices.

        Parameters
        ----------
        coin_in : str, int
            ID of coin to be priced; in a swapping context, this is
            the "in"-token.
        coin_out : str, int
            ID of quote currency; in a swapping context, this is the
            "out"-token.
        use_fee: bool, default=True
            Deduct fees.

        Returns
        -------
        float
            Price of `coin_in` quoted in `coin_out`
        """
        i, j = self.get_asset_indices(coin_in, coin_out)
        bp_token_index = self.n_total

        if bp_token_index not in (i, j):
            return self.dydx(i, j, use_fee=use_fee)

        i, j = self.get_meta_asset_indices(i, j, bp_token_index)
        xp = self._xp()
        return self._dydx(i, j, xp=xp, use_fee=use_fee)

    @override
    def trade(self, coin_in, coin_out, size):
        """
        Perform an exchange between two coins.

        Coin IDs should be strings but as a legacy feature integer indices
        corresponding to the pool implementation are allowed (caveat lector).

        Note that all amounts are normalized to be in the same units as
        pool value, e.g. for Curve Stableswap pools, the same units as `D`.
        This simplifies cross-token comparisons and creation of metrics.


        Parameters
        ----------
        coin_in : str, int
            ID of "in" coin.
        coin_out : str, int
            ID of "out" coin.
        size : int
            Amount of coin `i` being exchanged.

        Returns
        -------
        (int, int)
            (amount of coin `j` received, trading fee)
        """
        i, j = self.get_asset_indices(coin_in, coin_out)
        bp_token_index = self.n_total

        if bp_token_index not in (i, j):
            return self.exchange_underlying(i, j, size)

        i, j = self.get_meta_asset_indices(i, j, bp_token_index)
        return self.exchange(i, j, size)

    def get_meta_asset_indices(self, i, j, bp_token_index):
        """
        Get metapool asset indices from the output of get_coin_indices.
        """
        max_coin = self.max_coin

        if i == bp_token_index:
            i = max_coin

        if j == bp_token_index:
            j = max_coin

        if i == j:
            raise CurvesimValueError("Duplicate coin indices.")

        if i > max_coin or j > max_coin:
            raise CurvesimValueError(
                f"Index exceeds max metapool index (Input: {(i,j)}, Max: {max_coin})."
            )

        return i, j

    @override
    def get_max_trade_size(self, coin_in, coin_out, out_balance_perc=0.01):
        """
        Calculate the swap amount of the "in" coin needed to leave
        the specified percentage of the "out" coin.

        Parameters
        ----------
        coin_in : str, int
            ID of "in" coin.
        coin_out : str, int
            ID of "out" coin.
        out_balance_perc : float
            Percentage of the "out" coin balance that should remain
            after doing the swap.

        Returns
        -------
        int
            The amount of "in" coin needed.
        """
        i, j = self.get_asset_indices(coin_in, coin_out)

        max_coin = self.max_coin

        xp_meta = self._xp_mem(self.balances, self.rates)
        xp_base = self._xp_mem(self.basepool.balances, self.basepool.rates)

        base_i = i - max_coin
        base_j = j - max_coin
        meta_i = max_coin
        meta_j = max_coin
        if base_i < 0:
            meta_i = i
        if base_j < 0:
            meta_j = j

        if base_i < 0 or base_j < 0:
            xp_j = int(xp_meta[meta_j] * out_balance_perc)
            in_amount = self.get_y(meta_j, meta_i, xp_j, xp_meta)
            in_amount -= xp_meta[meta_i]
        else:
            xp_j = int(xp_base[base_j] * out_balance_perc)
            in_amount = self.basepool.get_y(base_j, base_i, xp_j, xp_base)
            in_amount -= xp_base[base_i]

        return in_amount

    @override
    def get_min_trade_size(self, coin_in):
        """
        Return the minimal trade size allowed for the pool.

        Parameters
        ----------
        coin_in : str, int
            ID of "in" coin.

        Returns
        -------
        int
            The minimal trade size
        """
        return 0

    @property
    @override
    @cache
    def assets(self):
        """
        Return :class:`.SimAssets` object with the properties of the pool's assets.

        Returns
        -------
        SimAssets
            SimAssets object that stores the properties of the pool's assets.
        """
        symbols = self.coin_names[:-1] + self.basepool.coin_names
        addresses = self.coin_addresses[:-1] + self.basepool.coin_addresses

        return SimAssets(symbols, addresses, self.chain)



from curvesim.exceptions import UnregisteredPoolError
from curvesim.pool.sim_interface import (
    SimCurveCryptoPool,
    SimCurveMetaPool,
    SimCurvePool,
    SimCurveRaiPool,
)


def get_pool_state(pool):
    """
    Returns pool state for the input pool. Functions for each pool type are
    specified in the `pool_state_functions` dict. Each function returns the
    values necessary to reconstruct pool state throughout a simulation run.
    """
    try:
        return pool_state_functions[type(pool)](pool)
    except KeyError as e:
        raise UnregisteredPoolError(
            f"State getter not implemented for pool type '{type(pool)}'."
        ) from e


def get_cryptoswap_pool_state(pool):
    """Returns pool state for stableswap non-meta pools."""
    return {
        "D": pool.D,
        "balances": pool.balances,
        "tokens": pool.tokens,
        "price_scale": pool.price_scale,
        "_price_oracle": pool._price_oracle,  # pylint: disable=protected-access
        "xcp_profit": pool.xcp_profit,
        "xcp_profit_a": pool.xcp_profit_a,
        "last_prices": pool.last_prices,
        "last_prices_timestamp": pool.last_prices_timestamp,
        "not_adjusted": pool.not_adjusted,
    }


def get_stableswap_pool_state(pool):
    """Returns pool state for stableswap non-meta pools."""
    return {
        "balances": pool.balances,
        "tokens": pool.tokens,
        "admin_balances": pool.admin_balances,
    }


def get_stableswap_metapool_state(pool):
    """Returns pool state for stableswap non-meta pools."""
    state = get_stableswap_pool_state(pool)
    bp_state = get_stableswap_pool_state(pool.basepool)
    bp_state = {key + "_base": val for key, val in bp_state.items()}
    state.update(bp_state)
    return state


pool_state_functions = {
    SimCurvePool: get_stableswap_pool_state,
    SimCurveMetaPool: get_stableswap_metapool_state,
    SimCurveRaiPool: get_stableswap_metapool_state,
    SimCurveCryptoPool: get_cryptoswap_pool_state,
}



"""
Simulation results object and function to initialize it from simulation pipeline output.
"""

__all__ = ["SimResults", "make_results"]

from .make_results import make_results
from .sim_results import SimResults



"""
Unit tests for CurveCryptoPool for n = 3

Tests are against the tricrypto-ng contract.
"""
import os
from itertools import permutations

import boa
from hypothesis import HealthCheck, assume, given, settings
from hypothesis import strategies as st

from curvesim.pool import CurveCryptoPool
from curvesim.pool.cryptoswap.calcs import get_p, get_y, newton_D
from curvesim.pool.cryptoswap.calcs.tricrypto_ng import (
    MAX_A,
    MAX_GAMMA,
    MIN_A,
    MIN_GAMMA,
    PRECISION,
    _newton_y,
    wad_exp,
)


def get_math(tricrypto):
    _base_dir = os.path.dirname(__file__)
    filepath = os.path.join(_base_dir, "../fixtures/curve/tricrypto_math.vy")
    _math = tricrypto.MATH()
    MATH = boa.load_partial(filepath).at(_math)
    return MATH


def initialize_pool(vyper_tricrypto):
    """
    Initialize python-based pool from the state variables of the
    vyper-based implementation.
    """
    A = vyper_tricrypto.A()
    gamma = vyper_tricrypto.gamma()
    n_coins = 3
    precisions = vyper_tricrypto.precisions()
    mid_fee = vyper_tricrypto.mid_fee()
    out_fee = vyper_tricrypto.out_fee()
    fee_gamma = vyper_tricrypto.fee_gamma()

    admin_fee = vyper_tricrypto.ADMIN_FEE()

    allowed_extra_profit = vyper_tricrypto.allowed_extra_profit()
    adjustment_step = vyper_tricrypto.adjustment_step()
    ma_half_time = vyper_tricrypto.ma_time()

    price_scale = [vyper_tricrypto.price_scale(i) for i in range(n_coins - 1)]
    price_oracle = [vyper_tricrypto.price_oracle(i) for i in range(n_coins - 1)]
    last_prices = [vyper_tricrypto.last_prices(i) for i in range(n_coins - 1)]
    last_prices_timestamp = vyper_tricrypto.last_prices_timestamp()
    balances = [vyper_tricrypto.balances(i) for i in range(n_coins)]
    # Use the cached `D`. See warning for `virtual_price` below
    D = vyper_tricrypto.D()
    lp_total_supply = vyper_tricrypto.totalSupply()
    xcp_profit = vyper_tricrypto.xcp_profit()
    xcp_profit_a = vyper_tricrypto.xcp_profit_a()

    pool = CurveCryptoPool(
        A=A,
        gamma=gamma,
        n=n_coins,
        precisions=precisions,
        mid_fee=mid_fee,
        out_fee=out_fee,
        allowed_extra_profit=allowed_extra_profit,
        fee_gamma=fee_gamma,
        adjustment_step=adjustment_step,
        admin_fee=admin_fee,
        ma_half_time=ma_half_time,
        price_scale=price_scale,
        price_oracle=price_oracle,
        last_prices=last_prices,
        last_prices_timestamp=last_prices_timestamp,
        balances=balances,
        D=D,
        tokens=lp_total_supply,
        xcp_profit=xcp_profit,
        xcp_profit_a=xcp_profit_a,
    )

    assert pool.A == vyper_tricrypto.A()
    assert pool.gamma == vyper_tricrypto.gamma()
    assert pool.balances == balances
    assert pool.price_scale == price_scale
    assert pool._price_oracle == price_oracle  # pylint: disable=protected-access
    assert pool.last_prices == last_prices
    assert pool.last_prices_timestamp == last_prices_timestamp
    assert pool.D == vyper_tricrypto.D()
    assert pool.tokens == lp_total_supply
    assert pool.xcp_profit == xcp_profit
    assert pool.xcp_profit_a == xcp_profit_a

    # WARNING: both `virtual_price` and `D` are cached values
    # so depending on the test, may not be updated to be
    # consistent with the rest of the pool state.
    #
    # Allowing this simplifies testing since we can avoid
    # coupling tests of basic functionality with the tests
    # for the complex newton calculations.
    #
    # We think it makes sense the initialized pool should
    # at least match the vyper pool (inconsistencies and all).
    # When full consistency is required, the `update_cached_values`
    # helper function should be utilized before calling
    # `initialize_pool`.
    virtual_price = vyper_tricrypto.virtual_price()
    pool.virtual_price = virtual_price

    return pool


def get_real_balances(virtual_balances, precisions, price_scale):
    """
    Convert from units of D to native token units using the
    given price scale.
    """
    assert len(virtual_balances) == 3
    balances = [x // p for x, p in zip(virtual_balances, precisions)]
    balances[1] = balances[1] * PRECISION // price_scale[0]
    balances[2] = balances[2] * PRECISION // price_scale[1]
    return balances


def update_cached_values(vyper_tricrypto):
    """
    Useful test helper after we manipulate the pool state.

    Calculates `D` and `virtual_price` from pool state and caches
    them in the appropriate storage.
    """
    A = vyper_tricrypto.A()
    gamma = vyper_tricrypto.gamma()
    xp = vyper_tricrypto.eval("self.xp()")
    xp = list(xp)  # boa doesn't like its own tuple wrapper
    MATH = get_math(vyper_tricrypto)
    D = MATH.newton_D(A, gamma, xp)  # pylint: disable=no-member
    vyper_tricrypto.eval(f"self.D={D}")
    total_supply = vyper_tricrypto.totalSupply()
    vyper_tricrypto.eval(
        f"self.virtual_price=10**18 * self.get_xcp({D})/{total_supply}"
    )


D_UNIT = 10**18
positive_balance = st.integers(min_value=10**5 * D_UNIT, max_value=10**11 * D_UNIT)
amplification_coefficient = st.integers(min_value=MIN_A, max_value=MAX_A)
gamma_coefficient = st.integers(min_value=MIN_GAMMA, max_value=MAX_GAMMA)
price = st.integers(min_value=10**12, max_value=10**25)


@given(
    st.integers(min_value=1, max_value=300),
    st.integers(min_value=0, max_value=2),
    st.integers(min_value=0, max_value=2),
)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_exchange(vyper_tricrypto, dx_perc, i, j):
    """Test `exchange` against vyper implementation."""
    assume(i != j)

    pool = initialize_pool(vyper_tricrypto)
    dx = pool.balances[i] * dx_perc // 100

    expected_dy = vyper_tricrypto.exchange(i, j, dx, 0)
    dy, _ = pool.exchange(i, j, dx)
    assert dy == expected_dy

    expected_balances = [vyper_tricrypto.balances(i) for i in range(3)]
    assert pool.balances == expected_balances


_num_iter = 10


@given(
    st.lists(
        st.integers(min_value=1, max_value=10000),
        min_size=_num_iter,
        max_size=_num_iter,
    ),
    st.lists(
        st.tuples(
            st.integers(min_value=0, max_value=2),
            st.integers(min_value=0, max_value=2),
        ).filter(lambda x: x[0] != x[1]),
        min_size=_num_iter,
        max_size=_num_iter,
    ),
    st.lists(
        st.integers(min_value=0, max_value=86400),
        min_size=_num_iter,
        max_size=_num_iter,
    ),
)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_multiple_exchange_with_repeg(
    vyper_tricrypto, dx_perc_list, indices_list, time_delta_list
):
    """Test `exchange` against vyper implementation."""

    pool = initialize_pool(vyper_tricrypto)

    for indices, dx_perc, time_delta in zip(
        indices_list, dx_perc_list, time_delta_list
    ):
        vm_timestamp = boa.env.vm.state.timestamp
        pool._block_timestamp = vm_timestamp

        i, j = indices
        dx = pool.balances[i] * dx_perc // 10000  # dx_perc in bps

        expected_dy = vyper_tricrypto.exchange(i, j, dx, 0)
        dy, _ = pool.exchange(i, j, dx)
        assert dy == expected_dy

        expected_balances = [vyper_tricrypto.balances(i) for i in range(3)]
        assert pool.balances == expected_balances

        assert pool.last_prices == [vyper_tricrypto.last_prices(i) for i in range(2)]
        assert pool.last_prices_timestamp == vyper_tricrypto.last_prices_timestamp()

        expected_price_oracle = [vyper_tricrypto.price_oracle(i) for i in range(2)]
        expected_price_scale = [vyper_tricrypto.price_scale(i) for i in range(2)]
        assert pool.price_oracle() == expected_price_oracle
        assert pool.price_scale == expected_price_scale

        boa.env.time_travel(time_delta)


@given(
    amplification_coefficient,
    gamma_coefficient,
    positive_balance,
    positive_balance,
    positive_balance,
)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_newton_D(vyper_tricrypto, A, gamma, x0, x1, x2):
    """Test D calculation against vyper implementation."""

    xp = [x0, x1, x2]
    assume(0.02 < xp[0] / xp[1] < 50)
    assume(0.02 < xp[1] / xp[2] < 50)
    assume(0.02 < xp[0] / xp[2] < 50)

    MATH = get_math(vyper_tricrypto)
    # pylint: disable=no-member
    expected_D = MATH.newton_D(A, gamma, xp)
    D = newton_D(A, gamma, xp)

    assert D == expected_D


@given(
    amplification_coefficient,
    gamma_coefficient,
    positive_balance,
    positive_balance,
    positive_balance,
)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=2,
    deadline=None,
)
def test_get_p(vyper_tricrypto, A, gamma, x0, x1, x2):
    """Test `get_p` calculation against vyper implementation."""

    xp = [x0, x1, x2]
    assume(0.02 < xp[0] / xp[1] < 50)
    assume(0.02 < xp[1] / xp[2] < 50)
    assume(0.02 < xp[0] / xp[2] < 50)

    MATH = get_math(vyper_tricrypto)
    # pylint: disable=no-member
    D = MATH.newton_D(A, gamma, xp)

    A_gamma = [A, gamma]
    expected_p = MATH.get_p(xp, D, A_gamma)
    p = get_p(xp, D, A, gamma)

    assert p == expected_p


@given(
    amplification_coefficient,
    gamma_coefficient,
    positive_balance,
    positive_balance,
    positive_balance,
    st.tuples(
        st.integers(min_value=0, max_value=2),
        st.integers(min_value=0, max_value=2),
    ).filter(lambda x: x[0] != x[1]),
    st.integers(min_value=1, max_value=10000),
)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_pure_get_y(vyper_tricrypto, A, gamma, x0, x1, x2, pair, dx_perc):
    """Test `get_y` calculation against vyper implementation."""
    i, j = pair

    xp = [x0, x1, x2]
    assume(0.02 < xp[0] / xp[1] < 50)
    assume(0.02 < xp[1] / xp[2] < 50)
    assume(0.02 < xp[0] / xp[2] < 50)

    MATH = get_math(vyper_tricrypto)
    # pylint: disable=no-member
    D = MATH.newton_D(A, gamma, xp)

    xp[i] += xp[i] * dx_perc // 10000

    expected_y_out = MATH.get_y(A, gamma, xp, D, j)
    y_out = get_y(A, gamma, xp, D, j)

    assert y_out[0] == expected_y_out[0]
    assert y_out[1] == expected_y_out[1]


def test_pool_get_y(vyper_tricrypto):
    """
    Test `CurveCryptoPool.get_y`.

    Note the pure version of `get_y` is already tested
    thoroughly in its own test against the vyper.

    This test is a sanity check to make sure we pass values in correctly
    to the pure `get_y` implementation.
    """
    pool = initialize_pool(vyper_tricrypto)

    xp = pool._xp()
    A = pool.A
    gamma = pool.gamma
    D = newton_D(A, gamma, xp)

    i = 0
    j = 1

    # `get_y` will set i-th balance to `x`
    x = xp[i] * 102 // 100
    y = pool.get_y(i, j, x, xp)

    xp[i] = x
    expected_y, _ = get_y(A, gamma, xp, D, j)

    assert y == expected_y


@given(st.integers(min_value=-42139678854452767551, max_value=135305999368893231588))
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=2,
    deadline=None,
)
def test_wad_exp(vyper_tricrypto, x):
    """Test the snekmate wad exp calc"""
    MATH = get_math(vyper_tricrypto)
    # pylint: disable=no-member
    expected_result = MATH.wad_exp(x)
    result = wad_exp(x)
    assert result == expected_result


@given(
    amplification_coefficient,
    gamma_coefficient,
    positive_balance,
    positive_balance,
    positive_balance,
    st.tuples(
        st.integers(min_value=0, max_value=2),
        st.integers(min_value=0, max_value=2),
    ).filter(lambda x: x[0] != x[1]),
    st.integers(min_value=1, max_value=10000),
)
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=1,
    deadline=None,
)
def test__newton_y(vyper_tricrypto, A, gamma, x0, x1, x2, pair, dx_perc):
    """Test D calculation against vyper implementation."""
    i, j = pair

    xp = [x0, x1, x2]
    assume(0.02 < xp[0] / xp[1] < 50)
    assume(0.02 < xp[1] / xp[2] < 50)
    assume(0.02 < xp[0] / xp[2] < 50)

    MATH = get_math(vyper_tricrypto)
    # pylint: disable=no-member
    D = MATH.newton_D(A, gamma, xp)

    xp[i] += xp[i] * dx_perc // 10000

    expected_y = MATH.eval(f"self._newton_y({A}, {gamma}, {xp}, {D}, {j})")
    y = _newton_y(A, gamma, xp, D, j)

    assert y == expected_y


def test_dydxfee(vyper_tricrypto):
    """Test spot price formula against execution price for small trades."""
    pool = initialize_pool(vyper_tricrypto)

    # USDT, WBTC, WETH
    decimals = [6, 8, 18]
    precisions = [10 ** (18 - d) for d in decimals]

    # print("WBTC price:", pool.price_scale[0] / 10**18)
    # print("WETH price:", pool.price_scale[1] / 10**18)

    dxs = [
        10**6,
        10**4,
        10**15,
    ]

    for pair in permutations([0, 1, 2], 2):
        i, j = pair

        dydx = pool.dydxfee(i, j)
        dx = dxs[i]
        dy = vyper_tricrypto.exchange(i, j, dx, 0)
        pool.exchange(i, j, dx, 0)  # update state to match vyper pool

        dx *= precisions[i]
        dy *= precisions[j]
        assert abs(dydx - dy / dx) / (dy / dx) < 1e-4



"""
Canonical source of truth for version info.

The version number should not be changed manually here.  Instead,
a tool such as `bump2version` should be used.
"""
import re

__version__ = "0.5.0.b1"


def parse_version(version):
    """
    Return a lexicographically comparable version sequence.
    """
    match = re.match(r"(\d+)\.(\d+)\.(\d+)\.?([a-z]*)(\d*)", version)
    major, minor, patch, release, build = match.groups()

    return (
        int(major),
        int(minor),
        int(patch),
        release or "release",  # '' means 'release'
        int(build) if build else 0,
    )


__version_info__ = parse_version(__version__)


if __name__ == "__main__":
    print(parse_version("0.3.5"))  # (0, 3, 5, 'release', 0)
    print(parse_version("0.3.5.release"))  # (0, 3, 5, release, 0)
    print(parse_version("0.3.5.rc3"))  # (0, 3, 5, rc, 3)
    print(parse_version("0.3.5.rc0"))  # (0, 3, 5, rc, 0)
    print(parse_version("0.3.5.a12"))  # (0, 3, 5, a, 12)



import asyncio

from ..network.subgraph import pool_snapshot_sync, symbol_address_sync
from ..network.web3 import underlying_coin_info_sync


def from_address(address, chain, env="prod", end_ts=None):
    """
    Returns

    Parameters
    ----------
    address: str
        Address prefixed with '0x'
    chain: str
        Chain name
    env: str
        Environment name for subgraph: 'prod' or 'staging'

    Returns
    -------
    Pool snapshot dictionary in the format returned by
    :func:`curvesim.network.subgraph.pool_snapshot`.
    """
    loop = asyncio.get_event_loop()
    data = pool_snapshot_sync(address, chain, env=env, end_ts=end_ts, event_loop=loop)

    # Get underlying token addresses
    if data["pool_type"] == "LENDING":
        u_addrs, u_decimals = underlying_coin_info_sync(
            data["coins"]["addresses"], event_loop=loop
        )

        m = data.pop("coins")
        names = [n[1:] for n in m["names"]]

        data["coins"] = {
            "names": names,
            "addresses": u_addrs,
            "decimals": u_decimals,
            "wrapper": m,
        }

    return data


def from_symbol(symbol, chain, env):
    address = symbol_address_sync(symbol, chain)

    data = from_address(address, chain, env)

    return data



from curvesim.logging import get_logger
from curvesim.pool.cryptoswap.calcs import newton_D

from .base import PoolMetaDataBase

logger = get_logger(__name__)


class CryptoswapMetaData(PoolMetaDataBase):
    """Specific implementation of the `PoolMetaDataInterface` for Cryptoswap."""

    def init_kwargs(self, balanced=True, balanced_base=True, normalize=True):
        data = self._dict

        kwargs = {
            "A": data["params"]["A"],
            "gamma": data["params"]["gamma"],
            "n": len(data["coins"]["names"]),
            "D": data["reserves"]["D"],
            "mid_fee": data["params"]["mid_fee"],
            "out_fee": data["params"]["out_fee"],
            "allowed_extra_profit": data["params"]["allowed_extra_profit"],
            "fee_gamma": data["params"]["fee_gamma"],
            "adjustment_step": data["params"]["adjustment_step"],
            "price_scale": data["params"]["price_scale"],
            "admin_fee": data["params"]["admin_fee"],
            "ma_half_time": data["params"]["ma_half_time"],
            "tokens": data["reserves"]["tokens"],
            "xcp_profit": data["params"]["xcp_profit"],
            "xcp_profit_a": data["params"]["xcp_profit_a"],
        }
        n = kwargs["n"]

        if not normalize:
            kwargs["precisions"] = [10 ** (18 - d) for d in data["coins"]["decimals"]]
        else:
            kwargs["precisions"] = [1] * n

        if not balanced:
            if normalize:
                coin_balances = data["reserves"]["by_coin"]
            else:
                coin_balances = data["reserves"]["unnormalized_by_coin"]
            kwargs["balances"] = coin_balances

        # Aug 15, 2023
        #
        # This is fallback logic for missing subgraph functionality.  Right now
        # the subgraph will return 0 for crypto pools with more than 2 coins.
        # ETA for fix is expected to be within few weeks.
        #
        # Using the staging subgraph should suffice for tricrypto-ng factory pools.
        if n == 3:
            if len(kwargs["price_scale"]) != 2:
                logger.warning("Price scale is missing.  Using ad-hoc setting.")
                coin_balances = data["reserves"]["by_coin"]
                numeraire_balance = coin_balances[0]
                price_scale = [
                    numeraire_balance * 10**18 // coin_balances[1],
                    numeraire_balance * 10**18 // coin_balances[2],
                ]
                kwargs["price_scale"] = price_scale

                A = kwargs["A"]
                gamma = kwargs["gamma"]
                xp = [
                    coin_balances[0],
                    coin_balances[0],
                    coin_balances[0],
                ]

                D = newton_D(A, gamma, xp)
                kwargs["D"] = D

        # Due to outstanding subgraph bug, we need to do something for
        # the missing value.
        if not kwargs["ma_half_time"]:
            kwargs["ma_half_time"] = 600

        return kwargs

    @property
    def coins(self):
        return self._dict["coins"]["addresses"]

    @property
    def coin_names(self):
        return self._dict["coins"]["names"]

    @property
    def n(self):
        return len(self._dict["coins"]["names"])



"""Unit tests for SimPool .prepare_for_run method."""

from hypothesis import HealthCheck, given, settings
from hypothesis import strategies as st
from pandas import DataFrame


def test_cryptoswap2_prepare_for_run_same_price(sim_curve_crypto_pool):
    """Test SimCurveCryptoPool.prepare_for_run with 2-coin pool, no price changes"""
    _test_cryptoswap_prepare_for_run_same_price(sim_curve_crypto_pool)


def test_cryptoswap3_prepare_for_run_same_price(sim_curve_tricrypto_pool):
    """Test SimCurveCryptoPool.prepare_for_run with 3-coin pool, no price changes"""
    _test_cryptoswap_prepare_for_run_same_price(sim_curve_tricrypto_pool)


@given(st.floats(min_value=0.1, max_value=0.9))
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_cryptoswap2_prepare_for_run_low_price(sim_curve_crypto_pool, scalar):
    """Test SimCurveCryptoPool.prepare_for_run with 2-coin pool, lower prices"""
    _test_cryptoswap_prepare_for_run(sim_curve_crypto_pool, scalar)


@given(st.floats(min_value=0.1, max_value=0.9))
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_cryptoswap3_prepare_for_run_low_price(sim_curve_tricrypto_pool, scalar):
    """Test SimCurveCryptoPool.prepare_for_run with 3-coin pool, lower prices"""
    _test_cryptoswap_prepare_for_run(sim_curve_tricrypto_pool, scalar)


@given(st.floats(min_value=1.1, max_value=10))
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_cryptoswap2_prepare_for_run_high_price(sim_curve_crypto_pool, scalar):
    """Test SimCurveCryptoPool.prepare_for_run with 2-coin pool, higher prices"""
    _test_cryptoswap_prepare_for_run(sim_curve_crypto_pool, scalar)


@given(st.floats(min_value=1.1, max_value=10))
@settings(
    suppress_health_check=[HealthCheck.function_scoped_fixture],
    max_examples=5,
    deadline=None,
)
def test_cryptoswap3_prepare_for_run_high_price(sim_curve_tricrypto_pool, scalar):
    """Test SimCurveCryptoPool.prepare_for_run with 3-coin pool, higher prices"""
    _test_cryptoswap_prepare_for_run(sim_curve_tricrypto_pool, scalar)


def _test_cryptoswap_prepare_for_run_same_price(pool):
    """Test that preserving price_scale doesn't change D or xcp(D)"""

    original_D = pool.D
    original_xcp = pool._get_xcp(pool.D)
    original_price_scale = pool.price_scale

    prices = [10**18 / p for p in pool.price_scale]
    prices = DataFrame([prices, prices])

    pool.prepare_for_run(prices)

    # Test that price attributes updated correctly
    _test_list_equality(pool.price_scale, original_price_scale)
    _test_list_equality(pool._price_oracle, original_price_scale)
    _test_list_equality(pool.last_prices, original_price_scale)

    # Test that D/xcp didn't change
    new_xcp = pool._get_xcp(pool.D)

    assert abs(pool.D - original_D) < 10**10
    assert abs(new_xcp - original_xcp) < 10**10


def _test_cryptoswap_prepare_for_run(pool, scalar):
    """Test that changing price_scale doesn't change xcp(D), but changes D"""

    original_D = pool.D
    original_xcp = pool._get_xcp(pool.D)
    new_price_scale = [int(p * scalar) for p in pool.price_scale]

    prices = [10**18 / p for p in new_price_scale]
    prices = DataFrame([prices, prices])

    pool.prepare_for_run(prices)

    # Test that price attributes updated correctly
    _test_list_equality(pool.price_scale, new_price_scale)
    _test_list_equality(pool._price_oracle, new_price_scale)
    _test_list_equality(pool.last_prices, new_price_scale)

    # Test that xcp didn't change but D did
    new_xcp = pool._get_xcp(pool.D)

    assert abs(pool.D - original_D) > 10**10
    assert abs(new_xcp - original_xcp) < 10**10


def _test_list_equality(list1, list2, tol=10**10):
    """Test that differences are < tol for all elements"""
    assert all([abs(val1 - val2) < tol for val1, val2 in zip(list1, list2)])



"""
This module provides functionality for preprocessing data for chart creation.

It contains utility functions for downsampling and preprocessing data, resampling data,
converting data to histograms, and creating histograms.
"""
from pandas import Grouper, Series, concat

from curvesim.exceptions import PlotError


def preprocess_data(data, config, factors, downsample=False):
    """
    Preprocesses data for chart creation, with an option to downsample.

    Parameters
    ----------
    data : pandas.DataFrame
        The data to preprocess.
    config : dict
        The plot configuration.
    factors : list
        The list of factors.
    downsample : bool, optional
        Whether to downsample the data. Defaults to False.

    Returns
    -------
    dict
        The preprocessed data dictionary.
    """
    if downsample:
        return downsample_data(data, config, factors)

    factor_data = data[factors]  # preserve unrounded factors
    data_out = data.round(6)
    data_out[factors] = factor_data
    return {"main": data_out}


def downsample_data(data, config, factors):
    """
    Downscales the data according to provided configurations and factors.

    Parameters
    ----------
    data : pandas.DataFrame
        The data to downsample.
    config : dict
        The plot configuration.
    factors : list
        The list of factors.

    Returns
    -------
    dict
        The downsampled data dictionary.

    Raises
    ------
    PlotError
        If no resample strategy is found in the configuration for a metric.
    """
    data_out = {}
    resample = {"run": "last"}
    for metric, cfg in config.items():
        rule = cfg.get("resample")

        if rule:
            resample[metric] = rule

        elif rule is False:
            data_out[metric] = data.pop(metric)

        elif cfg["style"].startswith("histogram"):
            d = concat([data[factors], data.pop(metric)], axis=1)
            data_out[metric] = to_histograms(d, factors)

        else:
            raise PlotError(f"No resample strategy found in {metric} config.")

    data_out["main"] = resample_data(data, factors, resample)
    return data_out


def resample_data(data, factors, fns):
    """
    Resamples the data based on the provided factors and resampling functions.

    Parameters
    ----------
    data : pandas.DataFrame
        The data to resample.
    factors : list
        The list of factors.
    fns : dict
        The resampling functions.

    Returns
    -------
    pandas.DataFrame
        The resampled data.
    """
    groups = factors + [Grouper(freq="1D", key="timestamp")]
    resampled = data.groupby(groups).agg(fns).round(6)
    return resampled.reset_index()


def to_histograms(data, factors):
    """
    Converts the data to histograms.

    Parameters
    ----------
    data : pandas.DataFrame
        The data to convert.
    factors : list
        The list of factors.

    Returns
    -------
    pandas.DataFrame
        The data converted to histograms.
    """
    if not factors:
        return make_histogram(data).reset_index()

    return data.groupby(factors).apply(make_histogram).reset_index()


def make_histogram(data, bins=500):
    """
    Creates a histogram from the data.

    Parameters
    ----------
    data : pandas.DataFrame
        The data to create a histogram from.
    bins : int, optional
        The number of bins for the histogram. Defaults to 500.

    Returns
    -------
    pandas.Series
        The histogram.
    """
    metric = data.iloc[:, -1]
    minimum = Series(0, index=[metric.min()])

    hist = metric.value_counts(bins=bins, sort=False, normalize=True)
    hist.index = hist.index.right
    hist = concat([minimum, hist])
    hist.index.name, hist.name = metric.name, "frequency"
    return hist.round(6)



"""Utlity functions for general usage in Curvesim."""
import functools
import inspect
import os
import re
import sys
from dataclasses import dataclass as _dataclass
from itertools import combinations

from dotenv import load_dotenv

from curvesim.exceptions import CurvesimException, MissingEnvVarError

load_dotenv()

is_py310 = sys.version_info.minor >= 10 or sys.version_info.major > 3
"""
Some utils require knowing the python version.  Right now it's only important to
distinguish between 3.10 and earlier versions.
"""

_NOT_VALUE = object()
"""
Dummy value for optional default arg
so that any value, including `None`,
can be set as a default.
"""


def get_env_var(var_name, default=_NOT_VALUE):
    """
    Retrieve environment variable.

    Parameters
    ----------
    var_name: str
        Name of the environment variable.
    default: object
        Value to return if env var is missing.

    Returns
    -------
    str
        Value of the environment variable.

    Raise
    -----
    curvesim.exception.MissingEnvVarError
        Raised if default is not set and env var is missing.
    """
    var_value = os.getenv(var_name)
    if var_value is None:
        if default is _NOT_VALUE:
            raise MissingEnvVarError(f"Could not get env var: '{var_name}'")
        return default

    return var_value


def cache(user_function, /):
    """
    Simple lightweight unbounded cache.  Sometimes called "memoize".

    Returns the same as lru_cache(maxsize=None), creating a thin wrapper
    around a dictionary lookup for the function arguments. Because it
    never needs to evict old values, this is smaller and faster than
    lru_cache() with a size limit.

    The cache is threadsafe so the wrapped function can be used in
    multiple threads.

    ----
    This isn't in functools until python 3.9, so we copy over the
    implementation as in:
    https://github.com/python/cpython/blob/3.11/Lib/functools.py#L648
    """
    return functools.lru_cache(maxsize=None)(user_function)


def override(method):
    """
    Method decorator to signify and check a method overrides a method
    in a super class.

    Implementation taken from https://stackoverflow.com/a/14631397/1175053
    """
    stack = inspect.stack()
    base_classes = re.search(r"class.+\((.+)\)\s*\:", stack[2][4][0]).group(1)

    # handle multiple inheritance
    base_classes = [s.strip() for s in base_classes.split(",")]
    if not base_classes:
        raise CurvesimException("override decorator: unable to determine base class")

    # stack[0]=overrides, stack[1]=inside class def'n, stack[2]=outside class def'n
    derived_class_locals = stack[2][0].f_locals

    # replace each class name in base_classes with the actual class type
    for i, base_class in enumerate(base_classes):

        if "." not in base_class:
            base_classes[i] = derived_class_locals[base_class]

        else:
            components = base_class.split(".")

            # obj is either a module or a class
            obj = derived_class_locals[components[0]]

            for c in components[1:]:
                assert inspect.ismodule(obj) or inspect.isclass(obj)
                obj = getattr(obj, c)

            base_classes[i] = obj

    if not any(hasattr(cls, method.__name__) for cls in base_classes):
        raise CurvesimException(
            f'Overridden method "{method.__name__}" was not found in any super class.'
        )
    return method


def get_pairs(arg):
    """
    Get sorted pairwise combinations of an iterable.
    Integer inputs are treated as range(int).

    Parameters
    ----------
    arg: iterable or int
        The iterable to produce pairs from, or an integer specifying the range
        to produce pairs from.

    Returns
    -------
    list
        Sorted pairwise combinations of the input.
    """
    if isinstance(arg, int):
        arg = range(arg)

    return list(combinations(arg, 2))


def dataclass(*args, **kwargs):
    """
    Slightly modified version of the standard library's `dataclass`.

    The modification is to allow the setting of slots on versions of
    python before 3.10.

    Right now, we just remove the `slots` kwarg if it exists, but in the
    future we can implement our own custom slots for old versions.
    """
    if "slots" in kwargs and not is_py310:
        del kwargs["slots"]

    return _dataclass(*args, **kwargs)



from curvesim.logging import get_logger
from curvesim.metrics.state_log import StateLog
from curvesim.pipelines.vol_limited_arb.trader import VolumeLimitedArbitrageur
from curvesim.templates import Strategy

logger = get_logger(__name__)


class VolumeLimitedStrategy(Strategy):
    """
    Computes and executes volume-limited arbitrage trades at each timestep.
    """

    trader_class = VolumeLimitedArbitrageur
    state_log_class = StateLog

    def __init__(self, metrics, vol_mult):
        """
        Parameters
        -----------
        metrics : List[Metric]
            A list of metrics used to evaluate the performance of the strategy.
        vol_mult : float or numpy.ndarray
            Value(s) multiplied by market volume to specify volume limits.

            Can be a scalar or vector with values for each pairwise coin combination.
        """
        super().__init__(metrics)
        self.vol_mult = vol_mult

    def _get_trader_inputs(self, sample):  # pylint: disable=too-few-public-methods
        volume_limits = _compute_volume_limits(sample, self.vol_mult)
        return sample.prices, volume_limits


def _compute_volume_limits(sample, vol_mult):
    prices = sample.prices
    volumes = sample.volumes

    limits = {key: volumes[key] * vol_mult[key] for key in volumes}
    reversed_limits = {(j, i): lim * prices[(i, j)] for (i, j), lim in limits.items()}
    return {**limits, **reversed_limits}



"""Unit tests for SimStableswapBase"""
import pytest
import itertools

from curvesim.pool.sim_interface.asset_indices import AssetIndicesMixin
from curvesim.utils import override

# pylint: disable=redefined-outer-name


class FakeSimPool(AssetIndicesMixin):
    """
    Fake implementation of a subclass of `CoinIndicesMixin`
    for testing purposes.
    """

    @property
    @override
    def asset_names(self):
        return ["SYM_0", "SYM_1", "SYM_2"]

    @property
    @override
    def _asset_balances(self):
        return [100, 200, 300]


@pytest.fixture(scope="function")
def sim_pool():
    """Fixture for testing derived class"""
    return FakeSimPool()


def test_asset_indices(sim_pool):
    """Test index conversion and getting"""
    # Example calling by symbol
    result = sim_pool.get_asset_indices("SYM_2", "SYM_0")
    assert result == [2, 0]

    names = sim_pool.asset_names
    name_sets = [
        list(itertools.permutations(names, r=i)) for i in range(1, len(names) + 1)
    ]
    for lst in name_sets:
        for name_set in lst:
            name_set = list(name_set)
            result = sim_pool.get_asset_indices(*name_set)
            assert result == [names.index(symbol) for symbol in name_set]

    # Example calling by index
    result = sim_pool.get_asset_indices(1, 2)
    assert result == [1, 2]

    indices = sim_pool.asset_indices.values()
    index_sets = [
        list(itertools.permutations(indices, r=i)) for i in range(1, len(indices) + 1)
    ]
    for lst in index_sets:
        for index_set in lst:
            index_set = list(index_set)
            result = sim_pool.get_asset_indices(*index_set)
            assert result == index_set

    assert sim_pool.asset_indices == {"SYM_0": 0, "SYM_1": 1, "SYM_2": 2}


def test_asset_balances(sim_pool):
    assert sim_pool.asset_balances == {"SYM_0": 100, "SYM_1": 200, "SYM_2": 300}



import os

from curvesim.iterators.param_samplers import ParameterizedPoolIterator
from curvesim.iterators.price_samplers import PriceVolume
from curvesim.metrics import init_metrics
from curvesim.metrics.results import make_results
from curvesim.pipelines import run_pipeline
from curvesim.pipelines.simple.strategy import SimpleStrategy
from curvesim.pool import get_sim_pool
from curvesim.pool.cryptoswap.pool import CurveCryptoPool

from ..common import DEFAULT_METRICS, DEFAULT_PARAMS, TEST_CRYPTO_PARAMS, TEST_PARAMS


def pipeline(  # pylint: disable=too-many-locals
    pool_address,
    chain,
    *,
    variable_params=None,
    fixed_params=None,
    test=False,
    end_ts=None,
    days=60,
    src="coingecko",
    data_dir="data",
    ncpu=None,
    env="prod",
):
    """
    Implements the simple arbitrage pipeline.  This is a very simplified version
    of :func:`curvesim.pipelines.vol_limited_arb.pipeline`.

    At each timestep, the pool is arbitraged as close to the prevailing market
    price as possible for the coin pair generating the largest arbitrage profit.

    Parameters
    ----------
    pool_address : str
        '0x'-prefixed string representing the pool address.

    chain: str
        Identifier for blockchain or layer2.  Supported values are:
        "mainnet", "arbitrum", "optimism", "fantom", "avalanche", "matic", "xdai"

    variable_params : dict, defaults to broad range of A/fee values
        Pool parameters to vary across simulations.
        keys: pool parameters, values: iterables of ints

        Example
        --------
        >>> variable_params = {"A": [100, 1000], "fee": [10**6, 4*10**6]}

    fixed_params : dict, optional
        Pool parameters set before all simulations.
        keys: pool parameters, values: ints

        Example
        --------
        >>> fixed_params = {"D": 1000000*10**18}

    test : bool, optional
        Overrides variable_params to use four test values:

        .. code-block::

            {"A": [100, 1000], "fee": [3000000, 4000000]}

    end_ts : int, optional
        End timestamp in Unix time.  Defaults to 30 minutes before midnight of the
        current day in UTC.

    days : int, default=60
        Number of days to pull price/volume data for.

    src : str, default="coingecko"
        Source for price/volume data: "coingecko" or "local".

    data_dir : str, default="data"
        relative path to saved price data folder

    ncpu : int, default=os.cpu_count()
        Number of cores to use.

    Returns
    -------
    :class:`~curvesim.metrics.SimResults`

    """
    ncpu = ncpu or os.cpu_count()
    fixed_params = fixed_params or {}

    default_params = DEFAULT_PARAMS.copy()
    for key in DEFAULT_PARAMS:
        if key in fixed_params:
            del default_params[key]

    variable_params = variable_params or DEFAULT_PARAMS

    pool = get_sim_pool(pool_address, chain, env=env, end_ts=end_ts)

    if test:
        fixed_params = {}
        if isinstance(pool, CurveCryptoPool):
            variable_params = TEST_CRYPTO_PARAMS
        else:
            variable_params = TEST_PARAMS

    sim_assets = pool.assets
    price_sampler = PriceVolume(
        sim_assets, days=days, end=end_ts, data_dir=data_dir, src=src
    )

    # pylint: disable-next=abstract-class-instantiated
    param_sampler = ParameterizedPoolIterator(pool, variable_params, fixed_params)

    _metrics = init_metrics(DEFAULT_METRICS, pool=pool)
    strategy = SimpleStrategy(_metrics)

    output = run_pipeline(param_sampler, price_sampler, strategy, ncpu=ncpu)
    results = make_results(*output, _metrics)
    return results
